{-----------------------------------------------------------------------------
 Unit Name: ConstTypesAndGlobals
 Project Name : MapR2toR2
 Author:    Angel Montesinos Amilibia
 December 2008- January 2009
 Purpose: Constants, types, global variables and  some routines used by
    other units
-----------------------------------------------------------------------------}
{

MapR2toR2, workshop for smooth maps from R^2 to R^2
Copyright (C) 2008-2009 Angel Montesinos Amilibia,
   University of Valencia, Spain


This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit ConstTypesAndGlobals;
interface
uses
  SysUtils, Windows, Graphics, Classes;

const
  pi = 3.1415926535897932384;

  kCreator = 'AMonAmi';
  kFilter = '.22m';  {Extension of the files read and produced by the program}
  kFileVersion0 = '0';
  kFileVersion1 = '1';

  kImgWidth = 460; {Of image panel}
  kImgWidthMinus1 = kImgWidth - 1;
  kImgWidthMinus2 = kImgWidth - 2;

  kDomWidth = 280; {Of domain panel}
  kDomWidthMinus1 = kDomWidth-1;
  kMaxFarAway = 10000; {It is supposed that a pixel with a greater distance to the origin
                        (0,0) cannot be visible in the window}

  kParamWidth = 200;  {Of parameter panel}
  kHalfParamWidth = kParamWidth div 2;
  kParamWidthMinus1 = kParamWidth-1;

  kDom = 1;
  kImg = 2;

  kAlmostZero = 1.0e-6;
  kAlmostZero2 = 1.0e-12;

  kLastCoordLine = 20; {Number of coordinate lines drawn in each direction minus one}


{Colors}
  cHorLines = $006666FF;  {for horizontal lines in domain and its images}
  cVertLines = $00FF6666; {for vertical lines in domain and its images}
  cReverse = $00C8C8C8;   {for both on the far part of the sphere}
  cSingLocus =$00333333;   {for the singular locus in domain and its image}
  cSingLocusShaded= $00999999; {the same but in the far part of the sphere}

{Miscellaneous}
  kReturn = char(13);
  kEnter = char(3);
  kEscape = char(27);
  kTab = char($09);
  kDot = '.';

  kTheError = 1511724757;

  {Used in the divided differences for the computation of
   derivatives}
  kIncr= 1/2048;
  kIncr2= kIncr*kIncr;
  kIncr4Points = 1/4096;
  kTwoIncr4Points = 2*kIncr4Points;
  kTwelveIncr4Points = 12*kIncr4Points;

type
  mapType=(Real, Complex); {types of maps distinguished by the program}

  ByteArray= array of Byte;

  Vector2 = array[1..2] of double;
  Vector3 = array[1..3] of double;
  Segment2 = array[1..2] of vector2;

  cVector2= array[1..2] of Vector2;
  cVector3= array[1..3] of Vector2;

  cMatrix22 = array[1..2] of cVector2;
  cMatrix32 = array[1..3] of cVector2;

  CurveNode2 = record
    key : Integer;
    pt, imgPt: vector2;
  end;

  CurveNode3 = record
    key : Integer;
    pt: vector3;
  end;

  Curve2 = record
    last: Integer;
    nodes: array of curveNode2;
  end;

  Curve3 = record
    last: Integer;
    nodes: array of curveNode3;
  end;

  CurveNode33 = record
    key: Integer;
    x: Vector2;
    pt, imgPt: Vector3;
  end;

  Curve33 = record
    last: Integer;
    nodes: array of CurveNode33;
  end;


  StereoSettings= record {settings for stereographic projection}
    r, r2,    {radius of the sphere and its square}
    sc: Double;  {radius of resulting projection after zooming}
    c: Vector2;  {center of the sphere}
    ctr: TPoint; {c0ordinates of the sphere center in panel}
    h,v,n: Vector3;  {'n', unit direction of visualization; 'h,v', orthonormal
                      basis or plane orthogonal to 'n'}
  end;

  R2ToR1Function = function(const par: Vector2): Double;
  R2ToR2Function= function(const par: Vector2): Vector2;
  R2ToR3Function= function(const par: Vector2): Vector3;
  R3ToR2Function= function(const x: Vector3; var back: Boolean): Vector2;
  PtBToR2Function= function(const x,y: Integer; back: Boolean): Vector2;
  Diff1R2ToR2Procedure= procedure(const par: vector2; var x,xu,xv: Vector2);
  R2ToBPtFunction= function(const x: Vector2; var back: Boolean): TPoint;

  MapR2R2Diff1 =  procedure (const par: Vector2; var x, xu, xv: Vector2) of object;

var
  singLocus: Curve33;
  c_Infty: Vector2 = (kTheError, 0);
  c_Zero: Vector2 = (0,0);
  c_One: Vector2 = (1,0);

  stsDom, stsImg: StereoSettings; {for stereographic projection of the domain or
                                  the image spaces}
function Area(x,y,z: Vector3): Vector3;
function cAdd(z,w: Vector2): Vector2;
procedure cChs(var z: Vector2);
function cDiv(z,w: Vector2): Vector2;
function cEq(z,w: Vector2): Boolean;
function cMod(z: Vector2): Double;
function cMod2(z: Vector2): Double;
function CMult(z,w: Vector2): Vector2;
function cSubtr(z,w: Vector2): Vector2;
function ClipLine(a,b: Vector2; director: Boolean; rt: TRect): vector2;
function Distance(x,y: Vector3): Double;
function ExtracTitle(pathAndName, filter: AnsiString): AnsiString;
function FormatFloatString(x: Double; totalStrLength: Integer): string;
procedure InitializeSphCenterRadiusAndViewPoint(typ: Byte);
function InnerProd(x,y: Vector3): Double;
function Norm(x: Vector3): Double;
procedure ReadStreamSomething(Stream: TStream; var smth; count: Integer);
function ReadStreamStr(Stream: TStream): string;
procedure ReportIOError(er: Integer);
procedure WriteStreamSomething(Stream: TStream; const smth; count: Integer);
procedure WriteStreamStr(Stream: TStream; Str: string);

implementation

uses
  Dialogs, Forms, Math,
  ResStrings;

function Area(x,y,z: Vector3): Vector3;
{two times the area vector of a triangle with vertexes 'x,y,z'}
begin
  result[1]:= (z[2]-x[2])*(y[3]-x[3]) - (z[3]-x[3])*(y[2]-x[2]);
  result[2]:= (z[3]-x[3])*(y[1]-x[1]) - (z[1]-x[1])*(y[3]-x[3]);
  result[3]:= (z[1]-x[1])*(y[2]-x[2]) - (z[2]-x[2])*(y[1]-x[1]);
end;

function cAdd(z,w: Vector2): Vector2;
{Complex sum. Gives  z+w}
begin
  result[1]:= z[1] + w[1];
  result[2]:= z[2] + w[2];
end;

procedure cChs(var z: Vector2);
{multiplies 'z' by -1}
begin
  z[1]:= -z[1];
  z[2]:= -z[2];
end;


function cEq(z,w: Vector2): Boolean;
{equality of complex numbers}
begin
  result:= (z[1] = w[1])  and  (z[2] = w[2]);
end;

function ClipLine(a,b: Vector2; director: Boolean; rt: TRect): vector2;
{given the rectangle 'rt', we consider the line t -->  a + t*b  if
 'director=True', or else the line  t-> a + t*(b-a). Then the function
 returns as (result[1],result[2])  the interval of t on which that line
 is inside 'rt'. If  result[1] > result[2], then the line does not cut 'rt' }
var
  t1, t2, t: Double;
  dir: Vector2;
  i: Integer;
begin
  result[1]:= 1;
  result[2]:= -1;
  if director then
    dir:= b
  else for i:= 1 to 2 do
    dir[i]:= b[i]-a[i];
  if (dir[1] <> 0) then begin
    result[1]:= (rt.left - a[1])/dir[1];
    result[2]:= (rt.right - a[1])/dir[1];
    if result[2] < result[1] then begin
      t:= result[2];
      result[2]:= result[1];
      result[1]:= t;
    end;
  end
  else if (a[1]< rt.left) or (a[1] > rt.right) then
    Exit;
  if (dir[2] <> 0) then begin
    t1:= (rt.left - a[1])/dir[1];
    t2:= (rt.right - a[1])/dir[1];
    if t2 < t1 then begin
      t:= t2;
      t2:= t1;
      t1:= t;
    end;
    result[1]:= Max(result[1],t1);
    result[2]:= Min(result[2],t2);
  end
  else if (a[1]< rt.left) or (a[1] > rt.right) then begin
    result[1]:= 1;
    result[2]:= -1;
  end;
end;

function cMod(z: Vector2): Double;
{gives the module of z, sqrt(z*c(z))}
begin
  result:= sqrt(sqr(z[1]) + sqr(z[2]));
end;

function cMod2(z: Vector2): Double;
{gives the squared module of z, z*c(z)}
begin
  result:= sqr(z[1]) + sqr(z[2]);
end;

function cMult(z,w: Vector2): Vector2;
{Complex multiplication gives z*w}
begin
  result[1]:= z[1]*w[1] - z[2]*w[2];
  result[2]:= z[1]*w[2] + z[2]*w[1];
end;

function cSubtr(z,w: Vector2): Vector2;
{Complex subtraction. Gives  z-w}
begin
  result[1]:= z[1] - w[1];
  result[2]:= z[2] - w[2];
end;

function cDiv(z,w: Vector2): Vector2;
{Complex division. Gives  z/w}
var
  aux: Double;
begin
  aux:= sqr(w[1]) + sqr(w[2]);
  if aux=0 then
    result:= c_Infty
  else begin
    aux:= 1/aux;
    result[1]:= (z[1]*w[1] + z[2]*w[2])*aux;
    result[2]:= (-z[1]*w[2] + z[2]*w[1])*aux;
  end;
end;

function Distance(x,y: Vector3): Double;
begin
  result:= sqrt( sqr(x[1]-y[1])+sqr(x[2]-y[2])+sqr(x[3]-y[3]));
end;

function ExtracTitle(pathAndName, filter: AnsiString): AnsiString;
{Extracts the name of the file without the extension   'filter'. Note
 that 'filter' must include the initial dot}
var
  ext: AnsiString;
  len: Integer;
begin
  result := ExtractFilename(pathAndName);
  len := Length(result);
  if len > 4 then begin
    ext := lowerCase(Copy(result, len - 3, 4));
    if ext =  kFilter then
      result := Copy(result, 1, len - 4);
  end;
end;

function FormatFloatString(x: Double; totalStrLength: Integer): string;
{formats a decimal number by giving it a total length of 'totalStrLength'
 characters}
  var
    sigDig: Integer;
begin
  result:= FloatToStrF(x, ffFixed, totalStrLength+2,15);
  sigDig:= Length(result);
  if sigDig > totalStrLength  then
    sigDig:= totalStrLength;
  result:= Copy(result,1,sigDig);
end;

procedure Initialize;
begin
  InitializeSphCenterRadiusAndViewPoint(kDom);
  stsDom.c[1]:= 0;
  stsDom.c[2]:= 0;
  stsImg:= stsDom;
  stsDom.ctr.x:= kDomWidth div 2;
  stsDom.ctr.y:= stsDom.ctr.x;
  stsDom.sc:= round(0.45*kDomWidth)/stsDom.r;
  stsImg.ctr.x:= kImgWidth div 2;
  stsImg.ctr.y:= stsImg.ctr.x;
  stsImg.sc:= round(0.45*kImgWidth)/stsImg.r;
end;

procedure InitializeSphCenterRadiusAndViewPoint(typ: Byte);
begin
  if typ= kDom then with stsDom do begin
    sc:= 0.45*kDomWidth;
    ctr.x:= kDomWidth div 2;
    ctr.y:= ctr.X;
    r:= 1;
    r2:= 1;
    c[1]:= 0;
    c[2]:= 0;
    h[1]:= 1;
    h[2]:= 0;
    h[3]:= 0;
    v[1]:= 0;
    v[2]:= 1;
    v[3]:= 0;
    n[1]:= 0;
    n[2]:= 0;
    n[3]:= -1;
  end
  else if typ= kImg then with stsImg do begin
    sc:= 0.45*kImgWidth;
    ctr.x:= kImgWidth div 2;
    ctr.y:= ctr.x;
    r:= 1;
    r2:= 1;
    c[1]:= 0;
    c[2]:= 0;
    h[1]:= 1;
    h[2]:= 0;
    h[3]:= 0;
    v[1]:= 0;
    v[2]:= 1;
    v[3]:= 0;
    n[1]:= 0;
    n[2]:= 0;
    n[3]:= -1;
  end
end;

function InnerProd(x,y: Vector3): Double;
begin
  result:= x[1]*y[1] + x[2]*y[2] + x[3]*y[3]; 
end;


function Norm(x: Vector3): Double;
begin
  result:= sqrt( sqr(x[1])+sqr(x[2])+sqr(x[3]));
end;

procedure finalize;
begin
  if High(singLocus.nodes) >= 0 then
    singLocus.nodes:= nil;
end;

function ReadStreamDouble(Stream: TStream): double;
{writes a double to the stream}
begin
  Result := Stream.Read(Result, SizeOf(Double));
end;

function ReadStreamInt(Stream: TStream): integer;
{returns an integer from the stream}
begin
  Stream.Read(Result, SizeOf(Integer));
end;

procedure ReadStreamSomething(Stream: TStream; var smth; count: Integer);
{The untypped variable 'smth' is filled with 'count' bytes read from the
stream}
var
  copied: Integer;
begin
  copied := Stream.Read(smth, count);
  if count <> copied then
    raise Exception.Create('');
end;

function ReadStreamStr(Stream: TStream): string;
{returns a string from the stream}
var
  StrLen: integer;
begin
  {get length of string}
   Stream.Read(StrLen, SizeOf(Integer));
  {set string to get memory}
  SetLength(Result, StrLen);
  {read characters}
  if StrLen > 0 then
    Stream.Read(Result[1], StrLen);
end;

procedure ReportIOError(er: Integer);
var
  e: Integer;
  m: AnsiString;
begin
  case er of
    -1: m := eUnknownMapType;
    -2: m := eBadFileType;
    -3: m := eFileNotAssigned;
    -4: m := eErrorOnWritingFile;
    -5: m := eErrorOnReadingFile;
    -6: m := eVersionNotSupported;
  else begin
      e := GetLastError;
      m := SysErrorMessage(e);
    end;
  end;
  ShowMessage(m);
end;


procedure WriteStreamSomething(Stream: TStream; const smth; count: Integer);
{Write to stream a lump of 'count' bytes read from an untypped variable 'smth'}
var
  copied: Integer;
begin
  copied := Stream.Write(smth, count);
  if copied <> count then
    raise Exception.Create('');
end;

procedure WriteStreamStr(Stream: TStream; Str: string);
{writes a string to the stream}
var
  StrLen: integer;
begin
  {get length of string}
  StrLen := Length(Str);
  {write length of string}
  Stream.Write(StrLen, SizeOf(Integer));
  {write characters}
  Stream.Write(Str[1], StrLen);
end;



initialization
  Initialize;

finalization
  Finalize;


end.
