{-----------------------------------------------------------------------------
 Unit Name: FMapR2R2
 Project Name : MapR2toR2
 Author:    Angel Montesinos Amilibia
 January 2003--October 2008
 Purpose: Main form of the program that mainly manages the interaction
          with the user
-----------------------------------------------------------------------------}
{

MapR2toR2, workshop for smooth maps from R^2 to R^2
Copyright (C) 2008-2009 Angel Montesinos Amilibia,
   University of Valencia, Spain


This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}

unit FMapR2R2;

interface

uses
  Windows, Forms, Graphics, Controls, Dialogs, SHELLApi, StdActns,
  Classes, ActnList, Menus, StdCtrls, ComCtrls, ExtCtrls, Buttons, Messages,
  JPEG, Types, SysUtils,
  ConstTypesAndGlobals, OCFunctDDDoubleSafe, FImplicitCurveLine;

type
  TMapR2R2 = class(TForm)
    Domain: TPaintBox;
    F1Edit: TEdit;
    F2Edit: TEdit;
    xuvLabel: TLabel;
    yuvLabel: TLabel;
    uMinEdit : TEdit;
    uMaxEdit : TEdit;
    vMinEdit : TEdit;
    vMaxEdit : TEdit;
    uLabel: TLabel;
    vLabel: TLabel;
    ActionList1 : TActionList;
    EditCut1 : TEditCut;
    EditCopy1 : TEditCopy;
    EditPaste1 : TEditPaste;
    SaveDialog1 : TSaveDialog;
    OpenDialog1 : TOpenDialog;
    Image: TPaintBox;
    MainMenu1 : TMainMenu;
    File1 : TMenuItem;
    Open1: TMenuItem;
    Save1: TMenuItem;
    SaveAs1: TMenuItem;
    Print1: TMenuItem;
    Edit1 : TMenuItem;
    Cut1 : TMenuItem;
    Copy1 : TMenuItem;
    Paste1 : TMenuItem;
    Exit1 : TMenuItem;
    PrintDialog1 : TPrintDialog;
    Manual1: TMenuItem;
    N2 : TMenuItem;
    Quit : TAction;
    InfoBar : TStatusBar;
    pLabel: TLabel;
    qLabel: TLabel;
    pMinEdit : TEdit;
    pMaxEdit : TEdit;
    qMinEdit : TEdit;
    qMaxEdit : TEdit;
    pName: TLabel;
    qName: TLabel;
    qValue: TEdit;
    pValue: TEdit;
    xLabel: TLabel;
    yLabel: TLabel;
    xMinEdit: TEdit;
    xMaxEdit: TEdit;
    yMinEdit: TEdit;
    yMaxEdit: TEdit;
    Geometry: TMenuItem;
    uName: TLabel;
    uValueEdit: TEdit;
    vValueEdit: TEdit;
    vName: TLabel;
    xName: TLabel;
    xValueEdit: TEdit;
    yValueEdit: TEdit;
    yName: TLabel;
    DrawButton: TBitBtn;
    N1: TMenuItem;
    RealFunctions1: TMenuItem;
    ComplexFunction1: TMenuItem;
    Singularlocus1: TMenuItem;
    Preferences1: TMenuItem;
    AdjusttoPanel1: TMenuItem;
    PreserveRatio1: TMenuItem;
    N3: TMenuItem;
    N4: TMenuItem;
    Planeimage1: TMenuItem;
    Stereographicimage1: TMenuItem;
    N5: TMenuItem;
    ResetsphereImg: TMenuItem;
    Parameters: TPaintBox;
    InfinityDomBox: TLabel;
    InfinityImgBox: TLabel;
    N6: TMenuItem;
    Planedomain1: TMenuItem;
    Stereographicdomain1: TMenuItem;
    ResetSphereDom: TMenuItem;
    procedure AdjustMenus(preferences: Integer);
    procedure AdjustSpotToParameters;
    procedure ChangeDomain(x,y: Integer; translation: Boolean);
    procedure ChangeImageBounds(x,y: Integer; translation: Boolean);
    procedure ChangeMapType(fromFile: Boolean);
    function CheckJump(p1,p2: Vector2): Boolean;
    procedure ComputeAndDrawCurves;
    procedure ComputeDomainCurves;
    procedure ComputeImageCurves;
    procedure ComputeSingularLocus;
    procedure CreateForm(Sender: TObject);
    procedure DestroyForm(Sender: TObject);
    procedure DrawCrossInImg(var dmPt, imgPt: Vector2; x,y: Integer);
    procedure DrawDomainCurves(rc: TRect; can: TCanvas);
    procedure DrawImageCurves(rc: TRect; can: TCanvas);
    procedure DrawCurves(typ: Byte);
    procedure DrawLittleCircle(pt: TPoint; toErase : Boolean);
    procedure DrawSingularLocusDom(rc: TRect; can: TCanvas);
    procedure DrawSingularLocusImg(rc: TRect; can: TCanvas);
    procedure EstimateImgBoundsAndDivisor;
    procedure ExecuteQuit(Sender : TObject);
    function FromParametersToParamPanel(pp: Vector2): TPoint;
    function FromParamPanelToParameters(pt: TPoint): Vector2;
    function GuessStep(var pp: vector2; var z: Vector3;  var step: Double;
                  minStep, maxStep: Double; coord: Integer): Boolean;
    procedure Inform(text: AnsiString);
    function NotFarAwayDom(x,y: Double): Boolean;
    function NotFarAwayImg(x,y: Double): Boolean;
    function NotFarAwayImgPanel(x,y: Double): Boolean;
    procedure OnCloseQuery(Sender : TObject; var CanClose : Boolean);
    procedure OnClosing(Sender : TObject; var Action : TCloseAction);
    procedure OnFunctionType(Sender: TObject);    
    procedure OnKeyDownInForm(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnKeyUpImageBounds(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure OnManual(sender: TObject);
    procedure OnMouseDownInDomain(Sender : TObject; Button : TMouseButton;
                  Shift : TShiftState; X, Y : Integer);
    procedure OnMouseDownInImage(Sender : TObject; Button : TMouseButton;
                  Shift : TShiftState; X, Y : Integer);
    procedure OnMouseDownInParameters(Sender : TObject; Button : TMouseButton;
                  Shift : TShiftState; X, Y : Integer);
    procedure OnMouseMoveInDomain(Sender : TObject; Shift : TShiftState; X, Y : Integer);
    procedure OnMouseMoveInImage(Sender: TObject; Shift: TShiftState; X,Y: Integer);
    procedure OnMouseMoveInParameters(Sender : TObject; Shift : TShiftState; X, Y : Integer);
    procedure OnMouseUpInDomain(Sender: TObject; Button: TMouseButton;
                 Shift: TShiftState; X, Y: Integer);
    procedure OnMouseUpInImage(Sender: TObject; Button: TMouseButton;
                               Shift: TShiftState; X, Y: Integer);
    procedure OnGeometryMenu(Sender: TObject);
    procedure OnPreferences(Sender: TObject);
    procedure OnPrint(Sender: TObject);
    procedure OnResetSphere(Sender: TObject);
    procedure Open(Sender : TObject);
    procedure PaintDomain(Sender : TObject);
    procedure PaintImage(Sender : TObject);
    procedure PaintParameters(Sender : TObject);
    procedure ReadAndDraw(sender: TObject);
    function ReadDataComplex : Integer;
    function ReadDataReal : Integer;
    function ReadFromDisk(pathAndName: TFilename): Boolean;
    procedure SetBasicData;
    procedure SetClip(can: TCanvas; rWhole, rClip: TRect);
    procedure SetCoordinateChangesParam;
    procedure SetUpComplexScene(fromFile: Boolean);
    procedure SetUpRealScene(fromFile: Boolean);
    procedure Save(Sender : TObject);
    procedure SaveAs(Sender : TObject);
    procedure StoreSingLocusNode(currentX: Vector2; key: Integer);
    procedure UpdateSingularLocus(typ: Byte);
    function VerifyImageBoundsEntries: Boolean;
    procedure WantToSave(Sender : TObject; var Action: TCloseAction);
    procedure WriteToDisk(pathAndName: TFilename);
  private
    minParam, maxParam, edgesDom, halfEdgesDom, edgesParam, ctrDom, ctrImg: Vector2;
    trackingDomDispl, trackingDomScal, trackingImgDispl, trackingImgScal: Boolean;
    spanDom, radDom, spanImg, crossSc, divisor, maxJumpDom, maxJumpImg: Double;
    panel2paramSc, gridStep: Vector2;
    prevSpot, prevPtInDom, prevPtInImg, firstMouseImg, lastMouseImg,
        firstMouseDom, lastMouseDom: TPoint;
    crossInImg, crossInDom: array[1..2,1..2] of TPoint;
    crossInImgBack: Boolean;
    dirty, alreadyParsed, alreadyDrawn,  thereIsLittleCircle, thereIsCrossInImg,
        thereIsCrossInDom, trackSphereDom, trackSphereImg: Boolean;
  public
    horCoordLinesDom, vertCoordLinesDom,
    horCoordLinesImg, vertCoordLinesImg: array[0..kLastCoordLine] of Curve3;

    minDom, maxDom, minImg, maxImg, edgesImg, halfEdgesImg: Vector2;
    mt: MapType;
    domBmp, imgBmp: TBitmap;
    domRct, imgRct, paramRct: TRect;
    p, q: Double;
    pComplex: Vector2;
    preserveRatio, wholeRange, modeStDomain, modeStImage,
       toDrawSingularLocus: Boolean;
    title: AnsiString;
    filename: TFileName;
  end;


var
  mapR2R2 : TMapR2R2;


implementation

{$R *.DFM}
uses
  Math, Printers,
  OCComplexFunct, ResStrings,
  PrintDrawings, FManMapR2R2, Maps;

const
  cxuvText = 'p u^2 - v^3';
  cyuvText= 'q v^2 - u^3';
  cwzText = 'z(1-p) + p cos(z)';
  cWindowColor = $00B0E8FF;

  cWholeRangeMask = 1;
  cToDrawSingularLocusMask = 2;
  cStImgMask = 4;
  cPreserveRatioMask = 8;
  cStDomMask = 16;


procedure TMapR2R2.AdjustMenus(preferences: Integer);
begin
  wholeRange:= (preferences and cWholeRangeMask) <> 0;
  toDrawSingularLocus:= (preferences and cToDrawSingularLocusMask) <> 0;
  modeStImage:= (preferences and cStImgMask) <> 0;
  modeStDomain:= (preferences and cStDomMask) <> 0;
  preserveRatio:= (preferences and cPreserveRatioMask) <> 0;
  SingularLocus1.checked:= toDrawSingularLocus;
  AdjustToPanel1.checked:= not preserveRatio;
  PreserveRatio1.checked:= preserveRatio;
  PlaneImage1.checked:= not modeStImage;
  StereographicImage1.checked:= modeStImage;
  PlaneDomain1.checked:= not modeStDomain;
  StereographicDomain1.checked:= modeStDomain;
end;

procedure TMapR2R2.AdjustSpotToParameters;
{Adjust parameter spot to the value of 'p', and when 'tg=2',
 adjust q-cursor to the value of 'q'}
var
  pt: TPoint;
begin
  if panel2paramSc[1] <> 0 then
    pt.x := round((p - minParam[1])/panel2paramSc[1])
  else
    pt.x := kHalfParamWidth;
  if panel2paramSc[2] <> 0 then
    pt.y := round((-q + maxParam[2])/panel2paramSc[2])
  else
    pt.Y:= kHalfParamWidth;
  DrawLittleCircle(pt, True);
end;

procedure TMapR2R2.ChangeDomain(x,y: Integer; translation: Boolean);
{scale the domain or translate it, and draw everything accordingly. The idea
 is to give the feeling that the curves and singular locus are translated
 without change when 'translation'= true. This means that the domain is to
 be shifted in the opposite sense. When 'translation'= False, the user changes
 the horizontal scale by moving the mouse to right and left, and
 the vertical by moving the mouse up and down.}
const
  cMul = 3;
var
  shX, shY, i: Integer;
  scX, scY: Extended;
begin
  if not alreadyDrawn then
    Exit;
  if translation then begin
    if not trackingDomDispl then begin
      trackingDomDispl:= True;
      trackingDomScal:= False;
      prevPtInDom.x := x;
      prevPtInDom.y := y;
    end;
  end
  else begin
    if not trackingDomScal then begin
      trackingDomDispl:= False;
      trackingDomScal:= True;
      prevPtInDom.x := x;
      prevPtInDom.y := y;
    end;
  end;
  if (x <> prevPtInDom.x) or (y <> prevPtInDom.y) then begin
    dirty:= True; {because there are new data to save}
    {compute shift of mouse}
    shX:= x - prevPtInDom.x;
    shY:= y - prevPtInDom.y;
    if translation then begin  {translate the domain}
      minDom[1]:= minDom[1] - shX * panel2domSc[1];
      minDom[2]:= minDom[2] + shY * panel2domSc[2];
      maxDom[1]:= maxDom[1] - shX * panel2domSc[1];
      maxDom[2]:= maxDom[2] + shY * panel2domSc[2];
      for i:=1 to 2 do
        ctrDom[i]:= (maxDom[i]+minDom[i])/2;
    end
    else begin
    {compute new scale and change the domain accordingly}
      scX:= (domRct.right-cMul*shx)/domRct.right;
      scY:= (domRct.bottom+cMul*shy)/domRct.bottom;
      maxDom[1]:= scX*halfEdgesDom[1] + ctrDom[1];
      minDom[1]:= ctrDom[1] - scX*halfEdgesDom[1];
      maxDom[2]:= scy*halfEdgesDom[2] + ctrDom[2];
      minDom[2]:= ctrDom[2] - scy*halfEdgesDom[2];
      for i:=1 to 2 do begin
        edgesDom[i]:= maxDom[i]-minDom[i];
        halfEdgesDom[i]:= edgesDom[i]/2;
      end;
    end;
      {update edit fields}
    uMinEdit.text:= FormatFloatString(minDom[1], 5 + ord(minDom[1] < 0));
    vMinEdit.text:= FormatFloatString(minDom[2], 5 + ord(minDom[2] < 0));
    uMaxEdit.text:= FormatFloatString(maxDom[1], 5 + ord(maxDom[1] < 0));
    vMaxEdit.text:= FormatFloatString(maxDom[2], 5 + ord(maxDom[2] < 0));
    update;
    SetBasicData;
    SetCoordinateChangesDom(DomRct, minDom, maxDom, preserveRatio);
    SetClip(domBmp.canvas, domRct, domEffRct);
    ComputeAndDrawCurves;
    prevPtInDom.x := x;
    prevPtInDom.y := y;
  end;
end;

procedure TMapR2R2.ChangeImageBounds(x,y: Integer; translation: Boolean);
{scale the image or translate it, and draw everything accordingly. The idea
 is to give the feeling that the curves and singular locus are translated
 without change when 'translation'= true. This means that the image bounds are to
 be shifted in the opposite sense. When 'translation'= False, the user changes
 the horizontal scale by moving the mouse to right and left, and
 the vertical by moving the mouse up and down.}
const
  cMul = 3;
var
  shX, shY, i: Integer;
  scX, scY: Extended;
begin
  if (not alreadyDrawn) or modeStImage then
    Exit;
  if translation then begin
    if not trackingImgDispl then begin
      trackingImgDispl:= True;
      trackingImgScal:= False;
      prevPtInImg.x := x;
      prevPtInImg.y := y;
    end;
  end
  else begin
    if not trackingImgScal then begin
      trackingImgDispl:= False;
      trackingImgScal:= True;
      prevPtInImg.x := x;
      prevPtInImg.y := y;
    end;
  end;
  if (x <> prevPtInImg.x) or (y <> prevPtInImg.y) then begin
    dirty:= True; {because there are new data to save}
    {compute shift of mouse}
    shX:= x - prevPtInImg.x;
    shY:= y - prevPtInImg.y;
    if trackingImgDispl then begin  {translate the image bounds}
      minImg[1]:= minImg[1] - shX * panel2ImgSc[1];
      minImg[2]:= minImg[2] + shY * panel2ImgSc[2];
      maxImg[1]:= maxImg[1] - shX * panel2ImgSc[1];
      maxImg[2]:= maxImg[2] + shY * panel2ImgSc[2];
      for i:=1 to 2 do
        ctrImg[i]:= (maxImg[i]+minImg[i])/2;
    end
    else if trackingImgScal then begin
    {compute new scale and change the image bounds accordingly}
      scX:= (ImgRct.right-cMul*shx)/ImgRct.right;
      scY:= (ImgRct.bottom+cMul*shy)/ImgRct.bottom;
      maxImg[1]:= scX*halfEdgesImg[1] + ctrImg[1];
      minImg[1]:= ctrImg[1] - scX*halfEdgesImg[1];
      maxImg[2]:= scy*halfEdgesImg[2] + ctrImg[2];
      minImg[2]:= ctrImg[2] - scy*halfEdgesImg[2];
      for i:=1 to 2 do begin
        edgesImg[i]:= maxImg[i]-minImg[i];
        halfEdgesImg[i]:= edgesImg[i]/2;
      end;
    end;
      {update edit fields}
    xMinEdit.text:= FormatFloatString(minImg[1], 5 + ord(minImg[1] < 0));
    yMinEdit.text:= FormatFloatString(minImg[2], 5 + ord(minImg[2] < 0));
    xMaxEdit.text:= FormatFloatString(maxImg[1], 5 + ord(maxImg[1] < 0));
    yMaxEdit.text:= FormatFloatString(maxImg[2], 5 + ord(maxImg[2] < 0));
    wholeRange:= False;
    update;
    SetBasicData;
    SetCoordinateChangesImg(ImgRct, minImg, maxImg, preserveRatio);
    SetClip(ImgBmp.canvas, ImgRct, ImgEffRct);
    ComputeAndDrawCurves;
    prevPtInImg.x := x;
    prevPtInImg.y := y;
  end;
end;

procedure TMapR2R2.ChangeMapType(fromFile: Boolean);
begin
  {initialize variables}
  trackSphereDom:= False;
  trackSphereImg:= False;
  thereIsCrossInDom := False;
  thereIsCrossInImg := False;

  {erase bitmaps and panels}
  domBmp.Canvas.FillRect(domEffRct);
  Domain.Canvas.FillRect(domEffRct);
  Image.Canvas.FillRect(imgeffRct);
  imgBmp.Canvas.FillRect(imgEffRct);
  Parameters.Canvas.FillRect(paramRct);

  if not fromFile then begin
    preserveRatio:= (mt =Complex); {by default, preserve ratio is true
         in case of complex functions, so that when they are holomorphic,
         the preservation of angles will be patent}
    adjustTopanel1.checked:= not preserveRatio;
    PreserveRatio1.checked:= preserveRatio;
    toDrawSingularLocus:= (mt = Real); {if the complex function is holomorphic,
     the singular set consists of isolated points, so that the locus will not
     be visible}
    SingularLocus1.Checked:= toDrawSingularLocus;
  end;

  alreadyParsed:= False;
  alreadyDrawn:= False;

  if (not fromFile) then begin
    {initialize parameter edit fields of the window}
    pValue.Text:= '0';
    pMinEdit.Text:= '-2';
    pMaxEdit.Text:= '2';
    qValue.Text:= '0';
    qMinEdit.Text:= '-2';
    qMaxEdit.Text:= '2';
  end;

  SingularLocus1.Enabled:= True;
  AdjustToPanel1.Enabled:= True;
  Parameters.Visible:= True;
  Parameters.Enabled:= True;
  pValue.visible:= True;
  pValue.enabled:= True;
  qValue.visible:= True;
  qValue.enabled:= True;
  pName.Visible:= True;
  qName.Visible:= True;
  pMinEdit.visible:= True;
  pMinEdit.enabled:= True;
  pMaxEdit.visible:= True;
  pMaxEdit.enabled:= True;
  pLabel.Visible:= True;
  qMinEdit.visible:= True;
  qMinEdit.enabled:= True;
  qMaxEdit.visible:= True;
  qMaxEdit.enabled:= True;
  qLabel.Visible:= True;
  F1Edit.ReadOnly := False;
  F1Edit.Color:= clWhite;
  ActiveControl:= TWinControl(F1Edit);
  F2Edit.Visible:= (mt = Real);
  F2Edit.Enabled:= F2Edit.Visible;
  yuvLabel.Visible:= F2Edit.Visible;

  case mt of
    Real: SetUpRealScene(fromFile);
    Complex: SetUpComplexScene(fromFile);
    else
      ;
  end;
end;

function TMapR2R2.CheckJump(p1,p2: Vector2): Boolean;
{'p1' and 'p2' are two points in the domain. This functions checks whether
the jump from the associate points }
var
 posDom, nextPosDom, posImg, nextPosImg: Vector3;
begin
  posDom:= DomToDomR3(p1);
  posImg:= DomToImgR3(p1);
  nextPosDom:= DomToDomR3(p2);
  nextPosImg:= DomToImgR3(p2);
  result:= (distance(PosDom, nextPosDom) < maxJumpDom ) and
                 (distance(PosImg, nextPosImg) < maxJumpImg )
end;

procedure TMapR2R2.ComputeAndDrawCurves;
//it is assumed that the functions and data have been already parsed
begin
  singLocus.last:= -1;
  Parameters.Canvas.FillRect(paramRct);
  thereIsLittleCircle:= False;
  AdjustSpotToParameters;
  SetCoordinateChangesDom(domRct, minDom, maxDom, preserveRatio);
  ComputeDomainCurves;
  ComputeImageCurves;
  SetCoordinateChangesImg(imgRct, minImg, maxImg, preserveRatio);
  SetClip(domBmp.canvas, domRct, domEffRct);
  SetClip(imgBmp.canvas, imgRct, imgEffRct);
  if toDrawSingularLocus then
    ComputeSingularLocus;
  DrawDomainCurves(domRct, domBmp.canvas);
  DrawImageCurves(imgRct, imgBmp.canvas);
  if toDrawSingularLocus then begin
    DrawSingularLocusDom(domRct, domBmp.canvas);
    DrawSingularLocusImg(imgRct, imgBmp.canvas);
  end;
  Domain.Canvas.Draw(0, 0, domBmp);
  Image.canvas.Draw(0, 0, imgBmp);
  alreadyDrawn:= True;
  thereIsCrossInImg := False;
  thereIsCrossInDom := False;
end;


procedure TMapR2R2.ComputeDomainCurves;
var
  v: Vector2;
  grStp, stp: Double;
  i, j, last: Integer;
begin
  SetClip(domBmp.canvas, domRct, domEffRct);
  if not modeStDomain then
    last:= 1
  else
    last:= 2*kLastCoordLine;
  grStp:= edgesDom[2]/kLastCoordLine;
  stp:= edgesDom[1]/last;
  //first, lines parallel to the x-axis
  for i:= 0 to kLastCoordLine do begin
    horCoordLinesDom[i].last:= last;
    if High(horCoordLinesDom[i].nodes) < last then
       SetLength(horCoordLinesDom[i].nodes, last+1);
    v[2]:= minDom[2]+ i*grStp;
    v[1]:= minDom[1];
    horCoordLinesDom[i].nodes[0].pt:= DomToDomR3(v);
    horCoordLinesDom[i].nodes[0].key:= 1;
    for j:= 1 to last do begin
      v[1]:= minDom[1] + j*stp;
      horCoordLinesDom[i].nodes[j].pt:= DomToDomR3(v);
      horCoordLinesDom[i].nodes[j].key:= 0;
    end;
  end;
    //and now, lines parallel to the y-axis}
  grStp:= edgesDom[1]/kLastCoordLine;
  stp:= edgesDom[2]/last;
  for i:= 0 to kLastCoordLine do begin
    vertCoordLinesDom[i].last:= last;
    if High(vertCoordLinesDom[i].nodes) < last then
      SetLength(vertCoordLinesDom[i].nodes, last+1);
    v[1]:= minDom[1]+ i*grStp;
    v[2]:= minDom[2];
    vertCoordLinesDom[i].nodes[0].pt:= DomToDomR3(v);
    vertCoordLinesDom[i].nodes[0].key:= 1;
    for j:= 1 to last do begin
      v[2]:= minDom[2] + j*stp;
      vertCoordLinesDom[i].nodes[j].pt:= DomToDomR3(v);
      vertCoordLinesDom[i].nodes[j].key:= 0;
    end;
  end;
end;

procedure TMapR2R2.ComputeImageCurves;
const
  prop= 60;
var
  nn, step, minStep, maxStep: Double;
  pp: Vector2;
  z, aux: Vector3;
  i, j, parts: Integer;
  toFinish, good, wr: Boolean;
begin
  if modeStImage then
    divisor:= stsImg.r/400
  else begin
    wr:= wholeRange;
    if wr then
      EstimateImgBoundsAndDivisor
    else
      divisor:= (abs(maxImg[1]-minImg[1]) + abs(maxImg[2]-minImg[2]))/800;
  end;
  minStep:= spanDom/(prop*kLastCoordLine);
  step:= minStep;
  for i:= 1 to 2 do
    gridStep[i]:= (maxDom[i]- minDom[i])/kLastCoordLine;
  maxStep:= gridStep[2]/3;
  for i:= 0 to kLastCoordLine do begin
    vertCoordLinesImg[i].last:= -1;
    toFinish:= False;
    SetLength(vertCoordLinesImg[i].nodes, 50);
    pp[1]:= minDom[1]+i*gridStep[1];
    pp[2]:= minDom[2];
    good:= GuessStep(pp,z,step,minStep,maxStep, 2); {obtain an estimation of the initial step}
    if not good then
      continue;

    repeat {we need to have two initial good points that will be used to
            determine the next step and so on}
      aux:= z;
      pp[2]:= pp[2]+ step;
      try
        z:= DomToImgR3(pp);
        vertCoordLinesImg[i].nodes[0].pt:= aux;
        vertCoordLinesImg[i].nodes[0].key:= 1;
        vertCoordLinesImg[i].nodes[1].pt:= z;
        vertCoordLinesImg[i].nodes[1].key:= 0;
        good:= True;
      except on Exception do
        good:= False;
      end;
    until good or (pp[2]+step >= maxDom[2]);
    if not good then
      continue;

    vertCoordLinesImg[i].last:= 1;
    pp[2]:= pp[2]+ step;
    repeat {begins the general iteration}
      try
        z:= DomToImgR3(pp);
        good:= NotFarAwayImg(z[1], z[2]);
      except on exception do begin
          good:= False;
          pp[2]:= pp[2]+ step/10;
          continue; {perhaps we can begin anew at the next point}
        end;
      end;
      nn:= Distance(z, vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last-1].pt);
      if nn <> 0 then begin
        aux:= Area(vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last-1].pt,
                   vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last].pt, z);
          {the quotient '|aux|/nn' is the distance of the vertex 'last'
           to the line containing the vertexes 'last-1' and 'z'}
        nn:= Norm(aux)/(nn*divisor);
        parts:= 1+ round(nn); {if the separation is too large, we shall divide
                              the step into several shorter steps}
      end
      else
        parts:= 1;
      if parts = 1 then begin
        inc(vertCoordLinesImg[i].last);
        if vertCoordLinesImg[i].last > high(vertCoordLinesImg[i].nodes)  then
          SetLength(vertCoordLinesImg[i].nodes, vertCoordLinesImg[i].last + 50);
        vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last].key:= ord(not good);
        {if we failed to get the image of the previous proposed point, then
         good = False, so the 'key' is equal to 1, meaning that at this point
         begins a new curve arc}

        vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last].pt:= z;
        good:= True; {because we have been able to compute this point}
      end
      else if parts > 1 then begin
        pp[2]:= pp[2]- step; {go back to the last computed domain point}
        step:= step/parts;
        if vertCoordLinesImg[i].last + parts > High(vertCoordLinesImg[i].nodes) then
          SetLength(vertCoordLinesImg[i].nodes, vertCoordLinesImg[i].last + parts + 50);
        for j:= 1 to parts do begin {compute intermediate points}
          pp[2]:= pp[2]+step;
          try
            z:= DomToImgR3(pp);
            good:= NotFarAwayImg(z[1], z[2]);
          except on exception do begin
              good:= False;
              continue;
            end;
          end;
          inc(vertCoordLinesImg[i].last);
          vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last].key:= ord(not good);
          good:= True; {because we have been able to compute this point}
          vertCoordLinesImg[i].nodes[vertCoordLinesImg[i].last].pt:= z;
        end;
      end;
      if parts = 1 then begin
        if nn > 0 then {perhaps we can increasee the step, but we
                        limit the increment conservatively}
          step:= Min(1.5*step, step/nn)
        else
          step:= 1.5*step;
      end
      else
        step:= 1.5*step; {note that this step is the previous step divided
                          by 'parts'. So, we are decreasing the overall
                          step a little}

      step:= Min(step,maxStep); {anyway a step greater than gridStep[2] is
                                     rejected}
      pp[2]:= pp[2]+ step;
      if (not toFinish) and (pp[2] > maxDom[2]) then begin
        pp[2]:= maxDom[2]; {so, in the next iteration the curve will end at the
                             end of the domain}
        toFinish:= True;
      end
    until pp[2] > maxDom[2];
  end;

  {Now we do the same with the other lines}
  maxStep:= gridStep[1]/3;
  for i:= 0 to kLastCoordLine do begin
    horCoordLinesImg[i].last:= -1;
    toFinish:= False;
    SetLength(horCoordLinesImg[i].nodes, 50);
    pp[2]:= minDom[2]+i*gridStep[2];
    pp[1]:= minDom[1];
    good:= GuessStep(pp,z,step, minStep, maxStep, 1);
    if not good then
      continue;
    repeat
      aux:= z;
      pp[1]:= pp[1]+ step;
      try
        z:= DomToImgR3(pp);
        horCoordLinesImg[i].nodes[0].pt:= aux;
        horCoordLinesImg[i].nodes[0].key:= 1;
        horCoordLinesImg[i].nodes[1].pt:= z;
        horCoordLinesImg[i].nodes[1].key:= 0;
        good:= True;
      except on Exception do
        good:= False;
      end;
    until good or (pp[1]+step >= maxDom[1]);
    if not good then
      continue;
    horCoordLinesImg[i].last:= 1;
    pp[1]:= pp[1]+ step;
    repeat
      try
        z:= DomToImgR3(pp);
        good:= NotFarAwayImg(z[1], z[2]);
      except on exception do begin
          good:= False;
          pp[1]:= pp[1]+ step/10;
          continue;
        end;
      end;
      nn:= Distance(z,horCoordLinesImg[i].nodes[horCoordLinesImg[i].last - 1].pt);
      if nn <> 0 then begin
        aux:= Area(horCoordLinesImg[i].nodes[horCoordLinesImg[i].last - 1].pt,
                   horCoordLinesImg[i].nodes[horCoordLinesImg[i].last].pt, z);
        nn:= Norm(aux)/nn;
        parts:= 1+ round(nn/divisor);
      end
      else
        parts:= 1;
      if parts = 1 then begin
        inc(horCoordLinesImg[i].last);
        if horCoordLinesImg[i].last > high(horCoordLinesImg[i].nodes)   then
          SetLength(horCoordLinesImg[i].nodes, horCoordLinesImg[i].last + 50);
        horCoordLinesImg[i].nodes[horCoordLinesImg[i].last].key:= ord(not good);
        good:= True;
        horCoordLinesImg[i].nodes[horCoordLinesImg[i].last].pt:= z;
      end
      else if parts > 1 then  begin
        pp[1]:= pp[1]- step;
        step:= step/parts;
        if horCoordLinesImg[i].last + parts > High(horCoordLinesImg[i].nodes) then
          SetLength(horCoordLinesImg[i].nodes, horCoordLinesImg[i].last +parts + 50);
        for j:= 1 to parts do begin
          pp[1]:= pp[1]+step;
          try
            z:= DomToImgR3(pp);
            good:= NotFarAwayImg(z[1], z[2]);
          except on exception do begin
              good:= False;
              continue;
            end;
          end;
          inc(horCoordLinesImg[i].last);
          horCoordLinesImg[i].nodes[horCoordLinesImg[i].last].key:= ord(not good);
          good:= True;
          horCoordLinesImg[i].nodes[horCoordLinesImg[i].last].pt:= z;
        end;
      end;
      if parts = 1 then begin
        if nn > 0 then
          step:= Min(1.5*step, step/nn)
        else
          step:= 1.5*step;
      end
      else
        step:= 1.5*step;
      step:= min(step, maxStep);
      pp[1]:= pp[1]+ step;
      if (not toFinish) and (pp[1] > maxDom[1]) then begin
        pp[1]:= maxDom[1];
        toFinish:= True;
      end
    until pp[1] > maxDom[1];
  end;
end;


procedure TMapR2R2.ComputeSingularLocus;
begin
  if modeStDomain then
    maxJumpDom:= stsDom.r/40
  else
    maxJumpDom:= spanDom/100;
  if modeStImage then
    maxJumpImg:= stsImg.r/40
  else
    maxJumpImg:= spanImg/100;
  singLocus.last:= -1;
  if toDrawSingularLocus then
    theImplicitCurve2D.ComputeImplicitCurve(theJac, theJacGrad, StoreSingLocusNode,
       CheckJump, minDom, maxDom, 50, 500, radDom)
end;

procedure TMapR2R2.CreateForm(Sender : TObject);
begin
  ForceCurrentDirectory:= True; {When opening a file, the directory of the
                        open dialog will be that where the program resides}
  SetRect(domRct,0,0,kDomWidth, kDomWidth);
  SetRect(imgRct, 0,0, kImgWidth, kImgWidth);
  SetRect(paramRct,0,0,kParamWidth, kParamWidth);

  {create bitmaps. The drawings are made in them. Afterward,
   their content is passed to the respective panels}
  domBmp:= TBitmap.Create;
  imgBmp:= TBitmap.Create;
  domBmp.width:= kDomWidth;
  domBmp.height:= kDomWidth;
  imgBmp.width:= kImgWidth;
  imgBmp.height:= kImgWidth;

  color:= cWindowColor; {window background color}

  InfinityDomBox.Color:= cWindowColor;
  InfinityImgBox.Color:= cWindowColor;

  crossSc:= 0.03;
  mt:= Real;

  thereIsCrossInImg := False;
  thereIsCrossInDom := False;
  preserveRatio:= False;
  wholeRange:= False;
  modeStImage:= False;
  modeStDomain:= False;
  toDrawSingularLocus:= True;
  thereIsLittleCircle:= False;
  prevSpot.x:= 100;
  prevSpot.Y:= 100;
  alreadyParsed:= False;
  alreadyDrawn:= False;
  trackingDomDispl:= False;
  trackingDomScal:= False;
  changeMapType(False);
  theF1:= TOCFunctDDDoubleSafe.Create;
  theF2:= TOCFunctDDDoubleSafe.Create;
  theImplicitCurve2D:= TImplicitCurveLine.create;
end;

procedure TMapR2R2.DestroyForm(Sender: TObject);
  var
    i: Integer;
begin
  domBmp.Free;
  imgBmp.Free;
  if High(singLocus.nodes) >= 0 then
    singLocus.nodes:= nil;
  for i:= 0 to kLastCoordLine do begin
     horCoordLinesDom[i].nodes:= nil;
     vertCoordLinesDom[i].nodes:= nil;
     horCoordLinesImg[i].nodes:= nil;
     vertCoordLinesImg[i].nodes:= nil;
  end;
  theF1.Free;
  theF2.Free;
  theImplicitCurve2D.Free;
end;

procedure TMapR2R2.DrawCrossInImg(var dmPt, imgPt: Vector2; x,y: Integer);
{Draw at the image 'f' by 'theF' of the point (h,v) of the Domain panel, a pair
 of segments, tangent to the image of the coordinate lines passing by the domain
 point. Return also in 'c' the domain point corresponding to (h,v)}
var
  sc: Double;
  vp: vector2;
  fu, fv, img3: Vector3;
  cr: array[1..2,1..2] of Vector3;
  can : TCanvas;
  pt: TPoint;
  i,j, bck: Integer;
  good: Boolean;
begin
  can:= Image.Canvas;
  can.Pen.Mode := pmNotXor;
  if thereIsCrossInImg then begin
    if crossInImgBack then
      can.Pen.Color:= $00888888
    else
      can.Pen.Color := clRed;
    can.moveto(crossInImg[1,1].x, crossInImg[1,1].y);
    can.lineto(crossInImg[1,2].x, crossInImg[1,2].y);
    can.moveto(crossInImg[2,1].x, crossInImg[2,1].y);
    can.lineto(crossInImg[2,2].x, crossInImg[2,2].y);
    thereIsCrossInImg:= False;
  end;
  bck:= 2;
  if InDom(dmPt,imgPt,x,y,bck) then begin
    crossInImgBack:= (bck = 1);
    sc:= crossSc*spanDom;  {scale for the size of the crossInImg}
    DomToImgR3diff1(dmPt,img3,fu,fv); {get the partial derivatives of 'theF'}
    vp:= ImgR3ToPanel(img3, crossInImgBack);
    pt.x:= round(vp[1]);
    pt.y:= round(vp[2]);
    good:= (pt.x > 0) and (pt.x < kImgWidth) and (pt.y > 0) and (pt.y < kImgWidth);
    {if the image point is inside the Image panel, compute crossInImg}
    if good then begin
      for i:= 1 to 3 do begin {compute the ends of the segments in R^2}
        cr[1,1,i]:= img3[i] + sc*fu[i];
        cr[1,2,i]:= img3[i] - sc*fu[i];
        cr[2,1,i]:= img3[i] + sc*fv[i];
        cr[2,2,i]:= img3[i] - sc*fv[i];
      end;
      try  {project the segments to the Image panel}
        for i:=1 to 2 do
          for j:= 1 to 2 do begin
            vp:= ImgR3ToPanel(cr[i,j], crossInImgBack);
            crossInImg[i,j].x:=  round(vp[1]);
            crossInImg[i,j].y:=  round(vp[2]);
          end;
      except on exception do
        Exit;
      end;
    end;
    if good then begin
      if crossInImgBack then
        can.Pen.Color:= $00888888
      else
        can.Pen.Color := clRed;
      can.moveto(crossInImg[1,1].x, crossInImg[1,1].y);
      can.lineto(crossInImg[1,2].x, crossInImg[1,2].y);
      can.moveto(crossInImg[2,1].x, crossInImg[2,1].y);
      can.lineto(crossInImg[2,2].x, crossInImg[2,2].y);
      thereIsCrossInImg:= True;
    end;
  end;
  can.Pen.Mode := pmCopy;
end;

procedure TMapR2R2.DrawCurves(typ: Byte);
begin
  if (typ and kDom <> 0) then begin
    domBmp.Canvas.FillRect(domEffRct);
    DrawDomainCurves(domRct, domBmp.canvas);
    if toDrawSingularLocus then
      DrawSingularLocusDom(domRct, domBmp.canvas);
    Domain.Canvas.Draw(0, 0, domBmp);
    thereIsCrossInDom := False;
  end;
  if (typ and kImg <> 0) then begin
    imgBmp.Canvas.FillRect(imgEffRct);
    DrawImageCurves(imgRct, imgBmp.canvas);
    if toDrawSingularLocus then
      DrawSingularLocusImg(imgRct, imgBmp.canvas);
    Image.canvas.Draw(0, 0, imgBmp);
    thereIsCrossInImg := False;
  end;
  alreadyDrawn:= True;
end;

procedure TMapR2R2.DrawDomainCurves(rc: TRect; can: TCanvas);
{draw evenly separated coordinate lines in the Domain panel, using
 different colors for each direction}
var
  v: Vector2;
  i,j,r: Integer;
  back, lastBack: Boolean;
begin
  SetCoordinateChangesDom(rc, minDom, maxDom, preserveRatio);
  can.Pen.Width:= 1 + round((rc.right-rc.left) div 500);

  if modeStDomain then begin
    {We draw the back segments in light gray}
       {first, lines parallel to the x-axis}
    can.Pen.Color:= cReverse;
    for i:= 0 to kLastCoordLine do begin
      v:= DomR3ToPanel(horCoordLinesDom[i].nodes[0].pt, back);
      can.MoveTo(round(v[1]), round(v[2]));
      for j:= 1 to horCoordLinesDom[i].last do begin
        lastBack:= back;
        v:= DomR3ToPanel(horCoordLinesDom[i].nodes[j].pt, back);
        if back or lastBack then
          can.LineTo(round(v[1]), round(v[2]))
        else
          can.MoveTo(round(v[1]), round(v[2]));
      end;
    end;

    {and now, lines parallel to the y-axis}
    for i:= 0 to kLastCoordLine do begin
      v:= DomR3ToPanel(vertCoordLinesDom[i].nodes[0].pt, back);
      can.MoveTo(round(v[1]), round(v[2]));
      for j:= 1 to vertCoordLinesDom[i].last do begin
        lastBack:= back;
        v:= DomR3ToPanel(vertCoordLinesDom[i].nodes[j].pt, back);
        if back or lastBack then
          can.LineTo(round(v[1]), round(v[2]))
        else
          can.MoveTo(round(v[1]), round(v[2]));
      end;
    end;
  end;

  {Now, the front segments in their colors}
  can.Pen.Color:= cHorLines;
  for i:= 0 to kLastCoordLine do begin
    v:= DomR3ToPanel(horCoordLinesDom[i].nodes[0].pt, back);
    can.MoveTo(round(v[1]), round(v[2]));
    for j:= 1 to horCoordLinesDom[i].last do begin
      lastBack:= back;
      v:= DomR3ToPanel(horCoordLinesDom[i].nodes[j].pt, back);
      if back or lastBack then
        can.MoveTo(round(v[1]), round(v[2]))
      else
        can.LineTo(round(v[1]), round(v[2]));
    end;
  end;

  {and now, lines parallel to the y-axis}
  can.Pen.Color:= cVertLines;
  for i:= 0 to kLastCoordLine do begin
    v:= DomR3ToPanel(vertCoordLinesDom[i].nodes[0].pt, back);
    can.MoveTo(round(v[1]), round(v[2]));
    for j:= 1 to vertCoordLinesDom[i].last do begin
      lastBack:= back;
      v:= DomR3ToPanel(vertCoordLinesDom[i].nodes[j].pt, back);
      if back or lastBack then
        can.MoveTo(round(v[1]), round(v[2]))
      else
        can.LineTo(round(v[1]), round(v[2]));
    end;
  end;
  if modeStDomain then begin
    can.Pen.Color:= clGray;
    can.Brush.Style:= bsClear;
    r:= round(stsDom.r*stsDom.sc);
    can.Ellipse(stsDom.ctr.x - r, stsDom.ctr.y - r,
              stsDom.ctr.x + r, stsDom.ctr.y + r);
    can.Brush.Style:= bsSolid;
  end;
end;

procedure TMapR2R2.DrawImageCurves(rc: TRect; can: TCanvas);
{draw evenly separated coordinate lines in the Domain panel, using
 different colors for each direction}
var
  v: Vector2;
  i,j,r: Integer;
  back, lastBack: Boolean;
begin
  SetCoordinateChangesImg(rc, minImg, maxImg, preserveRatio);
  can.Pen.Width:= 1 + round((imgRct.right-imgRct.left) div 500);

  if modeStImage then begin
    {We draw the back segments}
     {first, lines parallel to the x-axis}
    can.Pen.Color:= cReverse;
    for i:= 0 to kLastCoordLine do begin
      v:= imgR3ToPanel(horCoordLinesimg[i].nodes[0].pt, back);
      can.MoveTo(round(v[1]), round(v[2]));
      for j:= 1 to horCoordLinesimg[i].last do begin
        lastBack:= back;
        v:= imgR3ToPanel(horCoordLinesimg[i].nodes[j].pt, back);
        if horCoordLinesimg[i].nodes[j].key = 1 then
          can.MoveTo(round(v[1]), round(v[2]))
        else begin
          if back or lastBack then
            can.LineTo(round(v[1]), round(v[2]))
          else
            can.MoveTo(round(v[1]), round(v[2]));
        end;
      end;
    end;

     {and now, lines parallel to the y-axis}
    for i:= 0 to kLastCoordLine do begin
      v:= imgR3ToPanel(vertCoordLinesimg[i].nodes[0].pt, back);
      can.MoveTo(round(v[1]), round(v[2]));
      for j:= 1 to vertCoordLinesimg[i].last do begin
        lastBack:= back;
        v:= imgR3ToPanel(vertCoordLinesimg[i].nodes[j].pt, back);
        if vertCoordLinesimg[i].nodes[j].key = 1 then
          can.MoveTo(round(v[1]), round(v[2]))
        else begin
          if back or lastBack then
            can.LineTo(round(v[1]), round(v[2]))
          else
            can.MoveTo(round(v[1]), round(v[2]));
        end;
      end;
    end;
  end;


  {We draw the front segments}
  can.Pen.Color:= cHorLines;
  for i:= 0 to kLastCoordLine do begin
    v:= imgR3ToPanel(horCoordLinesimg[i].nodes[0].pt, back);
    can.MoveTo(round(v[1]), round(v[2]));
    for j:= 1 to horCoordLinesimg[i].last do begin
      lastBack:= back;
      v:= imgR3ToPanel(horCoordLinesimg[i].nodes[j].pt, back);
      if horCoordLinesimg[i].nodes[j].key = 1 then
        can.MoveTo(round(v[1]), round(v[2]))
      else begin
        if back or lastBack then
          can.MoveTo(round(v[1]), round(v[2]))
        else
          can.LineTo(round(v[1]), round(v[2]));
      end;
    end;
  end;

  {and now, lines parallel to the y-axis}
  can.Pen.Color:= cVertLines;
  for i:= 0 to kLastCoordLine do begin
    v:= imgR3ToPanel(vertCoordLinesimg[i].nodes[0].pt, back);
    can.MoveTo(round(v[1]), round(v[2]));
    for j:= 1 to vertCoordLinesimg[i].last do begin
      lastBack:= back;
      v:= imgR3ToPanel(vertCoordLinesimg[i].nodes[j].pt, back);
      if vertCoordLinesimg[i].nodes[j].key = 1 then
        can.MoveTo(round(v[1]), round(v[2]))
      else begin
        if back or lastBack then
          can.MoveTo(round(v[1]), round(v[2]))
        else
          can.LineTo(round(v[1]), round(v[2]));
      end;
    end;
  end;

  if modeStImage then begin
    can.Pen.Color:= clGray;
    can.Brush.Style:= bsClear;
    r:= round(stsImg.r*stsImg.sc);
    can.Ellipse(stsImg.ctr.x - r, stsImg.ctr.y - r,
       stsImg.ctr.x + r, stsImg.ctr.y + r);
    can.Brush.Style:= bsSolid;
  end;
end;

procedure TMapR2R2.DrawLittleCircle(pt: TPoint; toErase : Boolean);
{draw in the Parameters panel the little circle that marks the relative
 position of the current parameters value}
var
  oldColor: TColor;
begin
  with parameters.canvas do begin
    oldColor:= Pen.Color;
    Pen.Color:= clRed;
    Pen.Mode:= pmNotXor;
    if toErase and thereIsLittleCircle then begin
      {draw the circle in the previous position. Using the pmNotXoR
      mode, this erases the previous circle}
      moveto(prevSpot.x - 4, prevSpot.y + 1);
      lineto(prevSpot.x - 2, prevSpot.y + 3);
      lineto(prevSpot.x + 1, prevSpot.y + 3);
      lineto(prevSpot.x + 3, prevSpot.y + 1);
      lineto(prevSpot.x + 3, prevSpot.y - 2);
      lineto(prevSpot.x + 1, prevSpot.y - 4);
      lineto(prevSpot.x - 2, prevSpot.y - 4);
      lineto(prevSpot.x - 4, prevSpot.y - 2);
      lineto(prevSpot.x - 4, prevSpot.y + 1);
      thereIsLittleCircle:= False;
    end;

    {draw the circle in the current position. Using the pmNotXoR
     mode, this allows to erase the circle later}
    moveto(pt.x - 4, pt.y + 1);
    lineto(pt.x - 2, pt.y + 3);
    lineto(pt.x + 1, pt.y + 3);
    lineto(pt.x + 3, pt.y + 1);
    lineto(pt.x + 3, pt.y - 2);
    lineto(pt.x + 1, pt.y - 4);
    lineto(pt.x - 2, pt.y - 4);
    lineto(pt.x - 4, pt.y - 2);
    lineto(pt.x - 4, pt.y + 1);
    prevSpot:= pt;
    thereIsLittleCircle:=  True;
    Pen.Color := oldColor;
    Pen.Mode := pmCopy;
  end;
end;

procedure TMapR2R2.DrawSingularLocusDom(rc: TRect; can: TCanvas);
var
  v: Vector2;
  oldPen: TPen;
  i: Integer;
  back, lastBack, good, lastGood: Boolean;
begin
  if singLocus.last < 0 then
    Exit;
  SetCoordinateChangesDom(rc, minDom, maxDom, preserveRatio);
  oldPen:= can.Pen;
  can.Pen.Width:= 2 + round ((rc.right-rc.left) div 500);
  v:= DomR3ToPanel(singLocus.nodes[0].Pt,back);
  good:= hypot(v[1], v[2]) < kMaxFarAway;
  if good then
    can.MoveTo(round(v[1]),round(v[2]));
  if singLocus.last >= 1 then
    for i:= 1 to singLocus.last do begin
      lastBack:= back;
      lastGood:= good;
      v:= DomR3ToPanel(singLocus.nodes[i].Pt,back);
      good:= hypot(v[1], v[2]) < kMaxFarAway;
      if singLocus.nodes[i].key = 1 then begin
        if good then
          can.MoveTo(round(v[1]),round(v[2]));
      end
      else if good then begin
        if lastGood then begin
          if lastBack or back then
            can.Pen.Color:= cSingLocusShaded
          else
            can.Pen.Color:= cSingLocus;
          can.LineTo(round(v[1]),round(v[2]));
        end
        else
          can.MoveTo(round(v[1]),round(v[2]));
      end;
    end;
 can.Pen:= oldPen;
end;

procedure TMapR2R2.DrawSingularLocusImg(rc: TRect; can: TCanvas);
var
  v: Vector2;
  oldPen: TPen;
  i: Integer;
  back, lastBack, good, lastGood: Boolean;
begin
  if singLocus.last < 0  then
    Exit;
  //SetCoordinateChangesImg(rc, minImg, maxImg, preserveRatio);
  oldPen:= can.Pen;
  can.Pen.Width:= 2 + round ((rc.right-rc.left) div 500);

  v:= ImgR3ToPanel(singLocus.nodes[0].imgPt,back);
  good:= NotFarAwayImgPanel(v[1], v[2]);
  if good then
    can.MoveTo(round(v[1]),round(v[2]));
  if singLocus.last >= 1 then
    for i:= 1 to singLocus.last do begin
      lastBack:= back;
      lastGood:= good;
      v:= ImgR3ToPanel(singLocus.nodes[i].imgPt,back);
      good:= NotFarAwayImgPanel(v[1], v[2]);
      if singLocus.nodes[i].key = 1 then begin
        if good then
          can.MoveTo(round(v[1]),round(v[2]));
      end
      else begin
        if good then begin
          if lastGood then begin
            if lastBack or back then
              can.Pen.Color:= cSingLocusShaded
            else
              can.Pen.Color:= cSingLocus;
            can.LineTo(round(v[1]),round(v[2]));
          end
          else
            can.MoveTo(round(v[1]),round(v[2]));
        end
      end;
    end;
  can.Pen:= oldPen;
end;

procedure TMapR2R2.EstimateImgBoundsAndDivisor;
{Set the bounding rectangle of great part of the image
 and the 'divisor' that limits the length of the
 curve segments. See 'ComputeImageCurves'}
const
  pointsNo= 15;
var
  i, j, k: Integer;
  step, minC, maxC, x, y: Vector2;
begin
  for i:= 1 to 2 do begin
    minC[i]:= 1.0e20;
    maxC[i]:= -1.0e20;
    step[i]:= edgesDom[i]/pointsNo;
  end;
  for i:= 0 to pointsNo do begin
    x[1]:= minDom[1] + step[1]*i;
    for j:= 0 to pointsNo do begin
      x[2]:= minDom[2] + step[2]*j;
      try
        y:= theF(x);
        for k:= 1 to 2 do begin
          minC[k]:= Min(minC[k],y[k]);
          maxC[k]:= Max(maxC[k],y[k]);
        end;
      except on Exception do
        continue;
      end;
    end;
  end;
  for i:= 1 to 2 do begin
    minC[i]:= Max(minC[i],-1.0e6);
    maxC[i]:= Min(maxC[i], 1.0e6);
  end;
  minImg:= minC;
  maxImg:= maxC;
  for i:= 1 to 2 do begin
    ctrImg[i]:= (maxC[i] + minC[i])/2;

    {we will give more room for the drawing}
    halfEdgesImg[i]:= 0.55*(maxC[i] - minC[i]);
    edgesImg[i]:= 2*halfEdgesImg[i];
    minImg[i]:= ctrImg[i]-halfEdgesImg[i];
    maxImg[i]:= ctrImg[i]+halfEdgesImg[i];
  end;
  divisor:= (halfEdgesImg[1] + halfEdgesImg[2] )/400;
  spanImg:= hypot(edgesImg[1], edgesImg[2]);
end;

procedure TMapR2R2.ExecuteQuit(Sender : TObject);
begin
  close;
end;


function TMapR2R2.FromParametersToParamPanel(pp: Vector2): TPoint;
begin
  if panel2paramSc[1] <> 0 then
    result.x:= round((pp[1]-minParam[1])/panel2paramSc[1])
  else
    result.x:= kHalfParamWidth;
  if panel2paramSc[2] <> 0 then
    result.y:= round((-pp[2]+maxParam[2])/panel2paramSc[2])
  else
    result.y:= kHalfParamWidth;
end;

function TMapR2R2.FromParamPanelToParameters(pt: TPoint): Vector2;
begin
  result[1]:= pt.x*panel2paramSc[1] + minParam[1];
  result[2]:= -pt.y*panel2paramSc[2]+ maxParam[2];
end;


function TMapR2R2.GuessStep(var pp: vector2; var z: Vector3;
    var step: Double; minStep, maxStep: Double; coord: Integer): Boolean;
{computes the map in three points and then suggests the step to get a
 deviation of 'divisor' pixels starting from 'pp'. If the computation fails,
 it tries to get a new starting point that is returned in 'pp'. 'horizontal'
 is true if the step is to be taken in the sense of increasing abscissas; if false,
 of increasing ordinates. In 'z' is returned the value of 'FromDomainToSphere(pp)'.
 The computation assumes, as approximation, that the curve describes a circle
 with constant velocity}
var
  z1, z2: Vector3;
  p1, p2: Vector2;
  a, a2, b, b2, sinAlpha, sinBeta, m: Double;
begin
  result:= False;
  p1[3-coord]:= pp[3-coord];
  p2[3-coord]:= pp[3-coord];
  repeat
    try
      p1[coord]:= pp[coord] + kIncr;
      p2[coord]:= p1[coord] + kIncr;
      z:= DomToImgR3(pp);
      z1:= DomToImgR3(p1);
      z2:= DomToImgR3(p2);
      a2:= sqr((z[1]+z2[1])*0.5- z1[1])+sqr((z[2]+z2[2])*0.5- z1[2])+
             sqr((z[3]+z2[3])*0.5- z1[3]);
      a:= sqrt(a2);
      b2:= sqr(z[1]-z2[1]) + sqr(z[2]-z2[2]) + sqr(z[2]-z2[2]);
      b:= sqrt(b2);
      m:= a2+b2;
      if m > 1.0e-20 then begin
        sinAlpha:= sqrt(2)*a*b/m;
        if sinAlpha > 1.0e-7 then begin
          m:= 2*a*divisor/(a2+b2);
          sinBeta:= sqrt(abs(m*(2-m)));
          step:= sinBeta*kIncr/sinAlpha;
          step:= Min(step, gridStep[coord]);
          result:= True;
        end
        else begin
          step:= maxStep;
          result:= True;
        end
      end
      else
        raise exception.Create('');
    except on Exception do begin
        result:= False;
        pp[coord]:= pp[coord] + gridStep[coord]/20;
      end;
    end;
  until result or (pp[coord] >= maxDom[coord]);
  if result then
    step:= Min(maxStep, Max(minStep,step));
end;


procedure TMapR2R2.Inform(text : AnsiString);
{put 'text' in the status bar at the bottom of the window}
begin
  InfoBar.SimpleText:= text;
end;

function TMapR2R2.NotFarAwayDom(x,y: Double): Boolean;
begin
  if modeStDomain then
    result:= True
  else
    result:= abs(x)+abs(y)  <  30*spanDom;
end;

function TMapR2R2.NotFarAwayImg(x,y: Double): Boolean;
begin
  if modeStImage then
    result:= True
  else
    result:= abs(x)+abs(y)  <  30*spanImg;
end;

function TMapR2R2.NotFarAwayImgPanel(x,y: Double): Boolean;
begin
  if modeStImage then
    result:= True
  else
    result:= abs(x)+abs(y)  <  30*kImgWidth;
end;

procedure TMapR2R2.OnCloseQuery(Sender : TObject;
  var CanClose : Boolean);
var
  rslt : Word;
  Action : TCloseAction;
begin
  CanClose := True;
  if dirty then
    begin
      {Ask if user wants to save}
      rslt := MessageDlg(eSaveChanges + Caption + eBeforeExiting,
        mtConfirmation, [mbYes, mbNo, mbCancel], 0);
      if rslt = mrYes then
        begin
          WantToSave(Sender, Action); {open and manage save dialog}
          if Action = caNone then
            CanClose := False;
        end
      else if rslt = mrCancel then  {user cancelled}
        CanClose := False
      else
        CanClose := True;
    end;
end;

procedure TMapR2R2.OnClosing(Sender : TObject; var Action : TCloseAction);
{called when user clicks in the close icon}
begin
  Action := caFree; {proceed to close everything or to asking user if
                     the current map must be saved}
end;

procedure TMapR2R2.OnFunctionType(Sender: TObject);
{change type of map}
var
  t: MapType;
begin
  if (Sender = RealFunctions1 ) then
    t:= real
  else if (Sender =  ComplexFunction1) then
    t:= complex;
  if t <> mt then begin
    mt:= t;
    ChangeMapType(False);
  end;
end;

procedure TMapR2R2.OnGeometryMenu(Sender: TObject);
var
  oldDrSingLoc: Boolean;
begin
  oldDrSingLoc:= toDrawSingularLocus;
  SingularLocus1.Checked:= not SingularLocus1.checked;
  toDrawSingularLocus:= SingularLocus1.Checked;
  if (oldDrSingLoc <> toDrawSingularLocus) and alreadyDrawn then begin
    dirty:= True;
    ComputeSingularLocus;
    DrawCurves(kDom+kImg);
  end;
end;

procedure TMapR2R2.OnKeyDownInForm(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (key in [VK_UP,VK_DOWN,VK_LEFT]) then begin
    if thereIsCrossInImg then
      with Image.Canvas do begin  {erase crossInImg}
        Pen.Mode := pmNotXor;
        Pen.Color := clRed;
        moveto(crossInImg[1,1].x, crossInImg[1,1].y);
        lineto(crossInImg[1,2].x, crossInImg[1,2].y);
        moveto(crossInImg[2,1].x, crossInImg[2,1].y);
        lineto(crossInImg[2,2].x, crossInImg[2,2].y);
        Pen.Mode := pmCopy;
        thereIsCrossInImg:= False;
      end;
    {Size of the crosses is changed}
    case Key of
      VK_UP:
        crossSc:= min(50,crossSc*1.3); {increase size by 30%}
      VK_DOWN:
        crossSc:= max(1/50,crossSc/1.3); {decrease size by 30%}
      VK_LEFT:
        crossSc:= 1; {return to the default size}
      else
        ;
    end;
  end;
end;

procedure TMapR2R2.OnKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
{called when user has changed an edit field different from those
 for inputting the map. If the changed field is readable and consistent,
 redraw the map immediately}
var
  m : double;
  reDraw: Byte;
begin
  if not alreadyDrawn then
    Exit;
  reDraw:= 0;
  try
    ReadConstantField((Sender as TEdit).text, m);

    {in these cases, user has changed something relative to parameters.}
    if (sender = pValue) and (m <> p) then begin
      if (m < minParam[1]) then begin
        {new value of 'p' is less than current lower limit: change the limit}
        minParam[1] := m - (maxParam[1]-minParam[1]) / 2;
        pMinEdit.text := FloatToStrF(minParam[1], ffFixed, 4, 2);
      end
      else if (m > maxParam[1]) then begin
        {new value of 'p' is greater than current upper limit: change the limit}
        maxParam[1] := m + (maxParam[1]-minParam[1]) / 2;
        pMaxEdit.text := FloatToStrF(maxParam[1], ffFixed, 4, 2);
      end;
      p:= m;
      pComplex[1]:= m;
      reDraw:= 1;
    end
    else if (sender = qValue) and (m <> q) then begin
      if (m < minParam[2]) then begin
        minParam[2] := m - (maxParam[2]-minParam[2]) / 2;
        qMinEdit.text := FloatToStrF(minParam[2], ffFixed, 4, 2);
      end
      else if (m > maxParam[2]) then begin
        maxParam[2] := m + (maxParam[2]-minParam[2]) / 2;
        qMaxEdit.text := FloatToStrF(maxParam[2], ffFixed, 4, 2);
      end;
      q:= m;
      pComplex[2]:= m;
      SetCoordinateChangesParam;
      reDraw:= 1;
    end
    else if (Sender = pMinEdit) and (m <> minParam[1]) then begin
      {if new lower limit of 'p' is not lesser than 'p', push 'p' and
       upper limit up}
      if m >= p then begin
        redraw:= 1;
        if (m >= maxParam[1]) then begin
          maxParam[1] := m;
          pMaxEdit.Text:= FloatToStrF(maxParam[1], ffFixed, 4, 2);
        end;
        p:= m;
        pValue.Text:= FloatToStrF(p, ffFixed, 4, 2);
        pComplex[1]:= p;
      end;
      minParam[1]:= m;
      pMinEdit.Text:= FloatToStrF(minParam[1], ffFixed, 4, 2);
    end
    else if (Sender = pMaxEdit) and (m <> maxParam[1])  then begin
    {if new upper limit of 'p' is not greater than 'p', push 'p' and
       lower limit down}
      if m <= p then begin
        redraw:= 1;
        if (m <= minParam[1]) then begin
          minParam[1] := m;
          pMinEdit.Text:= FloatToStrF(minParam[1], ffFixed, 4, 2);
        end;
        p:= m;
        pValue.Text:= FloatToStrF(p, ffFixed, 4, 2);
        pComplex[1]:= p;
      end;
      maxParam[1]:= m;
      pMaxEdit.Text:= FloatToStrF(maxParam[1], ffFixed, 4, 2);
    end
    else if (Sender = qMinEdit) and (m <> minParam[2]) then begin
      {if new lower limit of 'q' is not lesser than 'q', push 'q' and
       upper limit up}
      if m >= q then begin
        redraw:= 1;
        if (m >= maxParam[2]) then begin
          maxParam[2] := m;
          qMaxEdit.Text:= FloatToStrF(maxParam[2], ffFixed, 4, 2);
        end;
        q:= m;
        qValue.Text:= FloatToStrF(q, ffFixed, 4, 2);
        pComplex[2]:= q;
      end;
      minParam[2]:= m;
      qMinEdit.Text:= FloatToStrF(minParam[2], ffFixed, 4, 2);
    end
    else if (Sender = qMaxEdit) and (m <> maxParam[2])  then begin
    {if new upper limit of 'q' is not greater than 'q', push 'q' and
       lower limit down}
      if m <= q then begin
        redraw:= 1;
        if (m <= minParam[2]) then begin
          minParam[2] := m;
          qMinEdit.Text:= FloatToStrF(minParam[2], ffFixed, 4, 2);
        end;
        q:= m;
        qValue.Text:= FloatToStrF(q, ffFixed, 4, 2);
        pComplex[2]:= q;
      end;
      maxParam[2]:= m;
      qMaxEdit.Text:= FloatToStrF(maxParam[2], ffFixed, 4, 2);
    end

    {user has changed the domain. If the new domain is consistent, redraw the map}
    else if sender = uMinEdit then begin
        redraw:= 2*ord( (m <> minDom[1]) and (m < maxDom[1]) );
        minDom[1]:= m;
    end
    else if Sender = uMaxEdit then begin
        redraw:= 2*ord( (m <> maxDom[1]) and (m > minDom[1]) );
        maxDom[1]:= m;
    end
    else if Sender = vMinEdit then begin
        redraw:= 2*ord( (m <> minDom[2]) and (m < maxDom[2]) );
        minDom[2]:= m;
    end
    else if Sender = vMaxEdit then begin
        redraw:= 2*ord( (m <> maxDom[2]) and (m > minDom[2]) );
        maxDom[2]:= m;
    end;
    if redraw <> 0 then begin
      SetBasicData;
      if redraw = 1 then begin
        if mt= real then begin
          theF1.UpdateParameters([p,q]);
          theF2.UpdateParameters([p,q]);
        end;
        SetCoordinateChangesParam;
        AdjustSpotToParameters;
      end;
      dirty := True; {because a field that could be saved has changed}
      ComputeAndDrawCurves;
    end;
  except on  E: Exception do
    if  (not (E is EAccessViolation)) and (not (E is EConvertError)) then
      Inform(E.Message);
  end;
end;



procedure TMapR2R2.OnKeyUpImageBounds(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if not alreadyDrawn then
    Exit;
  if VerifyImageBoundsEntries then begin
      wholeRange:= False;
      SetBasicData;
      ComputeAndDrawCurves;
      dirty:= True;
  end
  else if not wholeRange then begin
    wholeRange:= True;
    ComputeAndDrawCurves;
  end;
end;

procedure TMapR2R2.OnManual(sender: TObject);
{opens the user manual}
begin
  if  ((Manual<> nil) and Manual.Visible) then
    Exit;
  Manual:= TManual.Create(self);
  Manual.Show;
end;

procedure TMapR2R2.OnMouseDownInDomain(Sender : TObject; Button : TMouseButton;
                  Shift : TShiftState; X, Y : Integer);
begin
  if not alreadyDrawn then
    Exit;
  if (ssRight in Shift) then begin
    if modeStDomain then begin
     {begin tracking mouse for changing the modeStImage sphere}
      trackSphereDom:= True;
      firstMouseDom.x:= x;
      firstMouseDom.y:= y;
      lastMouseDom:= firstMouseDom
    end
  end;
end;

procedure TMapR2R2.OnMouseDownInImage(Sender : TObject; Button : TMouseButton;
                  Shift : TShiftState; X, Y : Integer);
begin
  if not alreadyDrawn then
    Exit;
  if (ssRight in Shift) then begin
    if modeStImage then begin
      {begin tracking mouse for changing the modeStImage sphere}
      trackSphereImg:= True;
      firstMouseImg.x:= x;
      firstMouseImg.y:= y;
      lastMouseImg:= firstMouseImg;
    end;
  end;
end;

procedure TMapR2R2.OnMouseDownInParameters(Sender : TObject;
  Button : TMouseButton; Shift : TShiftState; X, Y : Integer);
var
  v2: Vector2;
begin
  if alreadyDrawn then begin
    DrawLittleCircle(Point(x,y), True);

    {change parameter values and the corresponding edit fields}
    v2:= FromParamPanelToParameters(Point(x,y));
    p:= v2[1];
    q:= v2[2];
    pValue.text := FloatToStrF(p, ffFixed, 4, 4);
    qValue.text := FloatToStrF(q, ffFixed, 4, 4);
    pValue.update;
    qValue.update;

    {update parameters in functions}
    if mt = real then begin
      theF1.UpdateParameters([p,q]);
      theF2.UpdateParameters([p,q]);
    end
    else begin
      pComplex[1]:= p;
      pComplex[2]:= q;
    end;

    {redraw map}
    SetCoordinateChangesParam;
    dirty:= True;
    ComputeAndDrawCurves;
  end;
end;

procedure TMapR2R2.OnMouseMoveInDomain(Sender : TObject; Shift : TShiftState; X,
  Y : Integer);
var
  c, f: vector2;
begin
  if not alreadyDrawn then
    exit;
  if trackSphereDom then begin
    if MoveStereoSettings(stsDom, lastMouseDom, Point(x,y),Shift) then begin
          {recompute and draw}
      ComputeDomainCurves;
      if toDrawSingularLocus then
        UpdateSingularLocus(kDom);
      DrawCurves(kDom);
      lastMouseDom:= Point(x,y); {save mouse position}
    end;
  end
  else if (ssCtrl in Shift) then {translate domain intervals where the map is defined}
    ChangeDomain(x,y,True)
  else if (ssShift in Shift) then
    ChangeDomain(x,y,False)  {make domain intervals bigger or smaller}
  else begin
    trackingDomDispl:= False;
    trackingDomScal:= False;
    DrawCrossInImg(c, f, x, y);
    if thereIsCrossInImg then begin
      {update values for the point in domain and in image defined by the mouse}
      uValueEdit.Text:= FloatToStrF(c[1], ffFixed, 7, 4);
      vValueEdit.Text:= FloatToStrF(c[2], ffFixed, 7, 4);
      xValueEdit.Text:= FloatToStrF(f[1], ffFixed, 7, 4);
      yValueEdit.Text:= FloatToStrF(f[2], ffFixed, 7, 4);
    end;
  end
end;

procedure TMapR2R2.OnMouseMoveInImage(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
var
  pt: Vector2;
  back: Boolean;
begin
  if not alreadyDrawn then
    Exit;
  if trackSphereImg and MoveStereoSettings(stsImg, lastMouseImg, Point(x,y),Shift) then begin
          {recompute if necessary and draw}
    if (shift - [ssLeft, ssRight]) <> [] then begin
      ComputeImageCurves;
      if toDrawSingularLocus then
        ComputeSingularLocus;
    end
    else if toDrawSingularLocus then
      UpdateSingularLocus(kImg);
    DrawCurves(kImg);
    lastMouseImg:= Point(x,y); {save mouse position}
  end
  else if (ssCtrl in Shift) then {translate domain intervals where the map is defined}
    ChangeImageBounds(x,y,True)
  else if (ssShift in Shift) then
    ChangeImageBounds(x,y,False)  {make domain intervals bigger or smaller}
  else begin
    trackingImgDispl:= False;
    trackingImgScal:= False;
    if not modeStImage then begin {compute and show value of the
       image point defined by mouse position}
      pt:= ImgPanelToR2(x,y, back);
      xValueEdit.Text:= FloatToStrF(pt[1], ffFixed, 7, 4);
      yValueEdit.Text:= FloatToStrF(pt[2], ffFixed, 7, 4);
    end;
  end;
end;

procedure TMapR2R2.OnMouseMoveInParameters(Sender : TObject;
  Shift : TShiftState; X, Y : Integer);
var
  v2: Vector2;
begin
  if not alreadyDrawn then
    Exit;
  if (ssLeft in Shift) and ((x <> prevSpot.x) or (y <> prevSpot.y)) then
  if (x >= 0) and (x <= kParamWidth) and (y >= 0) and (y <= kParamWidth) then begin
    {change parameter values and the corresponding edit fields}
    v2:= FromParamPanelToParameters(Point(x,y));
    p:= v2[1];
    q:= v2[2];
    pValue.text := FloatToStrF(p, ffFixed, 4, 4);
    qValue.text := FloatToStrF(q, ffFixed, 4, 4);
    pValue.update;
    qValue.update;

    {update parameter values in functions}
    if mt = real then begin
      theF1.UpdateParameters([p,q]);
      theF2.UpdateParameters([p,q]);
    end
    else begin
      pComplex[1]:= p;
      pComplex[2]:= q;
    end;

    {recompute and draw}
    SetBasicData;
    SetCoordinateChangesParam;
    dirty:= True;
    ComputeAndDrawCurves;
  end;
end;

procedure TMapR2R2.OnMouseUpInDomain(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  trackSphereDom:= False;   {cease to track mouse for changing the sphere}
  trackingDomDispl:= False; {cease to track mouse when modifying the domain}
  trackingDomScal:= False;  {  id. id. }
end;

procedure TMapR2R2.OnMouseUpInImage(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  trackingImgDispl:= False; {cease to track mouse when modifying the domain}
  trackingImgScal:= False;  {  id. id. }
  trackSphereImg:= False; {cease to track mouse for changing the sphere}
end;

procedure TMapR2R2.OnPreferences(Sender: TObject);
var
  pr: Boolean;
begin
  if TMenuItem(Sender).tag < 2 then begin
    pr:= PreserveRatio1.checked <> preserveRatio;
    if alreadyDrawn and  pr then begin
      {toggle aspect ratio of presentation for domain and image}
      preserveRatio := not preserveRatio;
      SetCoordinateChangesDom(domRct, minDom, maxDom, preserveRatio);
      SetCoordinateChangesImg(imgRct, minImg, maxImg, preserveRatio);
      ComputeAndDrawCurves;
      dirty:= True;
    end;
  end
  else if TMenuItem(Sender).tag  < 4 then begin
    pr:= StereographicImage1.checked <> modeStImage;
    if pr then begin
      {toggle image presentation between plane and stereographic}
      modeStImage := not modeStImage;
      if alreadyParsed then
        UpdateMaps;
      if modeStImage then begin
        wholeRange:= True;
        imgEffRct:= imgRct;
      end;
      wholeRange:= not VerifyImageBoundsEntries;
      if alreadyDrawn then begin
         SetCoordinateChangesImg(imgRct, minImg, maxImg, preserveRatio);
         ComputeAndDrawCurves;
      end;
      dirty:= True;
    end
  end
  else if TMenuItem(Sender).tag  < 7 then begin
    pr:= StereographicDomain1.checked <> modeStDomain;
    if pr then begin
      {toggle domain presentation between plane and stereographic}
      modeStDomain := not modeStDomain;
      if alreadyParsed then begin
        UpdateMaps;
        if modeStDomain then
          domEffRct:= domRct;
        SetCoordinateChangesDom(domRct, minDom, maxDom, preserveRatio);
        ComputeDomainCurves;
        if toDrawSingularLocus then
          UpdateSingularLocus(kDom);
        DrawCurves(kDom);
      end;
      dirty:= True;
    end
  end;
end;

procedure TMapR2R2.OnPrint(Sender : TObject);
var
  n, i : Integer;
begin
  if not alreadyDrawn then
    Exit;
  if  PrintDialog1.Execute then begin
    SetPrinterRectangles;
    n:=  Printer.copies;
    if n >= 1 then begin
      Printer.BeginDoc;
      for i:= 1 to n do
        DrawOnPrinter;
      Printer.EndDoc;
    end;
  end;

  //redraw window after having been hidden by printing dialogs
  Domain.Invalidate;
  Image.Invalidate;
end;

procedure TMapR2R2.OnResetSphere(Sender: TObject);
{Reset sphere data with the default values and redraw if necessary}
begin
  if alreadyDrawn then begin
    if sender = ResetSphereImg then begin
      InitializeSphCenterRadiusAndViewPoint(kImg);
      ComputeImageCurves;
      if toDrawSingularLocus then
        updateSingularLocus(kImg);
      DrawCurves(kImg);
    end
    else begin
      InitializeSphCenterRadiusAndViewPoint(kDom);
      ComputeDomainCurves;
      if toDrawSingularLocus then
        updateSingularLocus(kDom);
      DrawCurves(kDom);
    end;
  end;
end;

procedure TMapR2R2.Open(Sender : TObject);
var
  Done : Boolean;
begin
  done := False;
  OpenDialog1.Options := [ofFileMustExist, ofHideReadOnly, ofNoChangeDir];
  while not Done do begin
    if OpenDialog1.Execute then begin
      if not (ofExtensionDifferent in OpenDialog1.Options) then
        filename:= OpenDialog1.FileName;
        if not ReadFromDisk(filename) then begin
          Inform(eErrorOnReadingFile); {something was wrong}
          alreadyDrawn:= False;
        end;
      done := True;
    end
    else { User cancelled }
      Done := True;
  end;
end;

procedure TMapR2R2.PaintDomain(Sender : TObject);
var
  pt: TPoint;
begin
  Domain.canvas.FillRect(domEffRct);
  if alreadyDrawn then begin
    Domain.Canvas.Draw(0, 0, domBmp);
    pt.x:= (crossInDom[1,1].x + crossInDom[1,2].x) div 2;
    pt.y:= crossInDom[1,1].y;
    if PtInRect(domain.canvas.ClipRect, pt) then
      thereIsCrossInDom:= False;
  end;
end;


procedure TMapR2R2.PaintImage(Sender : TObject);
var
  v: Vector2;
  back: Boolean;
  prevCr: TPoint;
begin
  Image.canvas.FillRect(imgEffRct);
  if alreadyDrawn then begin
    Image.Canvas.Draw(0, 0, imgBmp);
    try {compute the previous position of the crossInImg in Image panel}
      prevCr.x:= (crossInImg[1,1].x + crossInImg[1,2].x) div 2;
      prevCr.y:= crossInImg[1,1].y;
      v:= ImgR3ToPanel(DomToImgR3(DomPanelToR2(prevCr.x,prevCr.x, False)), back);
    except on exception do
      Exit;
    end;
    if PtInRect(Image.Canvas.ClipRect, Point(round(v[1]),round(v[2]))) then
     {previous crossInImg has been erased by the FillRect(imgEffRct) call}
      thereIsCrossInImg := False;
  end;
end;

procedure TMapR2R2.PaintParameters(Sender : TObject);
begin
  Parameters.canvas.FillRect(paramRct);
  if alreadyDrawn and
      PtInRect(Parameters.Canvas.ClipRect,prevSpot) then begin
    thereIsLittleCircle := False;
  end;
end;

procedure TMapR2R2.ReadAndDraw(sender: TObject);
{Read all data input by user and draw the map}
var
  rslt : Integer;
begin
  rslt:= 0;
  if mt = real then
    rslt:= ReadDataReal
  else if mt= complex then
    rslt:= ReadDataComplex;
  if rslt = 1 then begin
    {initializations}
    singLocus.last:= -1;
    Parameters.Canvas.FillRect(paramRct);
    thereIsLittleCircle:= False;
    AdjustSpotToParameters;
    UpdateMaps;
    SetCoordinateChangesDom(domRct, minDom, maxDom, preserveRatio);
    SetClip(domBmp.canvas, domRct, domEffRct);
    ComputeDomainCurves;
    ComputeImageCurves;
    SetCoordinateChangesImg(imgRct, minImg, maxImg, preserveRatio);
    if toDrawSingularLocus then
      ComputeSingularLocus;
    DrawCurves(kDom + kImg);
    alreadyDrawn:= True;

    {If the call was made after an open file, everything should be clean}
    if sender = DrawButton then
      dirty:= True;
  end;
  DrawButton.Default:= False;
end;

function TMapR2R2.ReadDataComplex: Integer;
 {If result = 0, there is something wrong. If result = 1, all was good}
var
  F1Text, F2Text, varText: ansiString;
begin
  Result:= 0;
  alreadyParsed:= False;
  alreadyDrawn:= False;

  {trim texts for eliminating blank spaces, tabs or line returns}
  F1Text:= Trim(F1Edit.Text);
  F2Text:= Trim(F2Edit.Text);
  try
    if (F1Text = '') then
      raise Exception.Create(eMissingFirstFunction);
    theComplexF.ParseNewText(F1Text, 'pz');

    varText := pMinEdit.Text;
    ReadConstantField(varText, minParam[1]);
    varText := pMaxEdit.Text;
    ReadConstantField(varText, maxParam[1]);
    if maxParam[1] < minParam[1] then
      raise Exception.Create(ePReMinLEthanMax);

    varText := qMinEdit.Text;
    ReadConstantField(varText, minParam[2]);
    varText := qMaxEdit.Text;
    ReadConstantField(varText, maxParam[2]);
    if maxParam[2] < minParam[2] then
      raise Exception.Create(ePImMinLEthanMax);

    varText := pValue.Text;
    ReadConstantField(varText, p);
    if (p < minParam[1]) or (p > maxParam[1]) then
      raise Exception.Create(epBtweenBounds);
    pComplex[1]:= p;

    varText := qValue.Text;
    ReadConstantField(varText, q);
    if (q < minParam[2]) or (q > maxParam[2]) then
      raise Exception.Create(eqBtweenBounds);
    pComplex[2]:= q;

    varText := uMinEdit.Text;
    ReadConstantField(varText, minDom[1]);
    varText := uMaxEdit.Text;
    ReadConstantField(varText, maxDom[1]);
    if maxDom[1] <= minDom[1] then
      raise Exception.Create(ezReMinLessThanMax);

    varText := vMinEdit.Text;
    ReadConstantField(varText, minDom[2]);
    varText := vMaxEdit.Text;
    ReadConstantField(varText, maxDom[2]);
    if maxDom[2] <= minDom[2] then
      raise Exception.Create(ezImMinLessThanMax);

    try
      if xMinEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(xMinEdit.Text), minImg[1]);
      if xMaxEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(xMaxEdit.Text), maxImg[1]);
      if yMinEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(yMinEdit.Text), minImg[2]);
      if yMaxEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(yMaxEdit.Text), maxImg[2]);
      {if any of these readings fails, flow passes to "except" below}

      {check if data are consistent}
      if maxImg[1] <= minImg[1] then
          raise Exception.Create(ewReMinLessThanMax);
      if maxImg[2] <= minImg[2] then
          raise Exception.Create(ewImMinLessThanMax);

      {data are consistent}
      wholeRange:= False;
      dirty:= True;
      Inform(eWholeRange);
    except on  E: Exception do begin
        {if the error was different from a usual reading error, jump
         to inform about it}
        if (not (E is EAccessViolation)) and (not (E is EConvertError)) then
          raise;
        wholeRange:= True;
        InfoBar.SimpleText:= '';
        InfoBar.Update;
         end
    end;
    SetBasicData;
    SetCoordinateChangesParam;
    AdjustSpotToParameters;
    Result:= 1;
    alreadyParsed := True;
  Except on E: Exception do begin
      InfoBar.SimpleText := E.message;
      AlreadyParsed := False;
    end;
  end;
end;

function TMapR2R2.ReadDataReal : Integer;
 {Read all data input by user and draw the map.
  Results in 0 if something was wrong. Results in 1 if all was well}
var
  F1Text, F2Text, varText: ansiString;
begin
  Result:= 0;
  alreadyParsed:= False;
  alreadyDrawn:= False;
  F1Text:= Trim(F1Edit.Text);
  F2Text:= Trim(F2Edit.Text);
  try
    if (F1Text = '') then
      raise Exception.Create(eMissingFirstFunction);
    theF1.ParseNewText(F1Text, 'uv', 'pq');
    theF1.BuildFirstDerivative;
    theF1.BuildSecondDerivative;

    if (F2Text = '') then
      raise Exception.Create(eMissingSecondFunction);
    theF2.ParseNewText(F2Text, 'uv', 'pq');
    theF2.BuildFirstDerivative;
    theF2.BuildSecondDerivative;

    varText := pMinEdit.Text;
    ReadConstantField(varText, minParam[1]);
    varText := pMaxEdit.Text;
    ReadConstantField(varText, maxParam[1]);
    if maxParam[1] < minParam[1] then
      raise Exception.Create(epMinLEthanMax);

    varText := qMinEdit.Text;
    ReadConstantField(varText, minParam[2]);
    varText := qMaxEdit.Text;
    ReadConstantField(varText, maxParam[2]);
    if maxParam[2] < minParam[2] then
      raise Exception.Create(eqMinLEthanMax);

    varText := pValue.Text;
    ReadConstantField(varText, p);
    if (p < minParam[1]) or (p > maxParam[1]) then
      raise Exception.Create(epBtweenBounds);

    varText := qValue.Text;
    ReadConstantField(varText, q);
    if (q < minParam[2]) or (q > maxParam[2]) then
      raise Exception.Create(eqBtweenBounds);

    theF1.UpdateParameters([p,q]);
    theF2.UpdateParameters([p,q]);

    varText := uMinEdit.Text;
    ReadConstantField(varText, minDom[1]);
    varText := uMaxEdit.Text;
    ReadConstantField(varText, maxDom[1]);
    if maxDom[1] <= minDom[1] then
      raise Exception.Create(euMinLessThanuMax);

    varText := vMinEdit.Text;
    ReadConstantField(varText, minDom[2]);
    varText := vMaxEdit.Text;
    ReadConstantField(varText, maxDom[2]);
    if maxDom[2] <= minDom[2] then
      raise Exception.Create(evMinLessThanvMax);

   try
      if xMinEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(xMinEdit.Text), minImg[1]);
      if xMaxEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(xMaxEdit.Text), maxImg[1]);
      if yMinEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(yMinEdit.Text), minImg[2]);
      if yMaxEdit.Text = '' then
        raise EConvertError.Create('');
      ReadConstantField(Trim(yMaxEdit.Text), maxImg[2]);
      {if any of these readings fails, flow passes to "except" below}

      {check if data are consistent}
      if maxImg[1] <= minImg[1] then
        raise Exception.Create(exMinLessThanxMax);
      if maxImg[2] <= minImg[2] then
        raise Exception.Create(eyMinLessThanyMax);

      wholeRange:= False;
      dirty:= True;
      Inform(eWholeRange);
    except on  E: Exception do begin
        {if the error was different from a usual reading error, jump
         to inform about it}
        if (not (E is EAccessViolation)) and (not (E is EConvertError)) then
          raise;
        wholeRange:= True;
        InfoBar.SimpleText:= '';
      end
    end;
    SetBasicData;
    SetCoordinateChangesParam;
    Result:= 1;
    alreadyParsed := True;

  Except on E: Exception do begin
      InfoBar.SimpleText := E.message;
      AlreadyParsed := False;
    end;
  end;
end;

function TMapR2R2.ReadFromDisk(pathAndName: TFilename): Boolean;
var
  mst: TMemoryStream;
  str: String;
  isMobius: Boolean;
  er, anInt, preferences, vers: Integer;
  anInt64: Int64;
begin
  isMobius:= False;
  mst := TMemoryStream.Create;
  er:= 0;
{$O-} {There are some pitfalls if optimization is enabled. Disable it}
  try
    mst.LoadFromFile(pathAndName);
    mst.Seek(0, soFromBeginning);
    str:= ReadStreamStr(mst);
    if str <> kCreator then
      er:= -2
    else begin
      str:= ReadStreamStr(mst);
      if str <> kFilter then
        er:= -2
      else begin
        str:= ReadStreamStr(mst);
        if str = kFileVersion0 then
          vers:= 0

        {added in kFileVersion1}
        else if str = kFileVersion1 then
          vers:= 1

        else
          er:= -6;

        if er = 0 then begin
          mst.Read(anInt64,SizeOf(int64)); {currently I haven't a use for this}
          mst.Read(anInt, SizeOf(Integer));
          case anInt of
            0: mt:= Real;
            1: mt:= Complex;
            2: begin
                 mt:= Complex;
                 isMobius:= True; {added in kFileVersion1 and corrected afterwards}
               end;
            else
              er:= -1;
          end;
          if er = 0 then try
            mst.Read(preferences, SizeOf(Integer));
            F1Edit.Text:= ReadStreamStr(mst);
            F2Edit.Text:= ReadStreamStr(mst);
            ReadStreamSomething(mst, minParam, SizeOf(Vector2));
            ReadStreamSomething(mst, maxParam, SizeOf(Vector2));
            mst.Read(p,SizeOf(Double));
            mst.Read(q,SizeOf(Double));
            ReadStreamSomething(mst, minDom, SizeOf(Vector2));
            ReadStreamSomething(mst, maxDom, SizeOf(Vector2));
            ReadStreamSomething(mst, minImg, SizeOf(Vector2));
            ReadStreamSomething(mst, maxImg, SizeOf(Vector2));
            ReadStreamSomething(mst, stsImg.c, SizeOf(Vector2));
            ReadStreamSomething(mst, stsImg.h, SizeOf(Vector3));
            ReadStreamSomething(mst, stsImg.v, SizeOf(Vector3));
            ReadStreamSomething(mst, stsImg.n, SizeOf(Vector3));
            mst.Read(stsImg.r, SizeOf(Double));

            {Added in fhe kFileVersion1}
            if vers = 1 then begin
              ReadStreamSomething(mst, stsDom, SizeOf(stsDom));
              ReadStreamSomething(mst, stsImg, SizeOf(stsImg));
              //ReadStreamSomething(mst, mm, SizeOf(mm));
            end;
            if isMobius then
              F1Edit.Text:='(p z + 1)/(z - 1)';

          except on exception do
            if er <> 0 then
            ReportIOError(er);
          end;
        end;
      end;
    end;
  finally
    mst.Free;
    if er <> 0 then
      ReportIOError(er);
{$O+} {Enable optimizations}
  end;
  result := (er = 0);
  if result then begin
    AdjustMenus(preferences);
    uMinEdit.Text:= FormatFloatString(minDom[1], 6);
    uMaxEdit.Text := FormatFloatString(maxDom[1], 6);
    vMinEdit.Text := FormatFloatString(minDom[2], 6);
    vMaxEdit.Text := FormatFloatString(maxDom[2], 6);
    pValue.Text:= FormatFloatString(p, 6);
    qValue.Text:= FormatFloatString(q, 6);
    pMinEdit.Text := FormatFloatString(minParam[1], 6);
    pMaxEdit.Text := FormatFloatString(maxParam[1], 6);
    qMinEdit.Text := FormatFloatString(minParam[2], 6);
    qMaxEdit.Text := FormatFloatString(maxParam[2], 6);
    if wholeRange then begin
      xMinEdit.Text := '';
      xMaxEdit.Text := '';
      yMinEdit.Text := '';
      yMaxEdit.Text := '';
    end
    else begin
      xMinEdit.Text := FormatFloatString(minImg[1], 6);
      xMaxEdit.Text := FormatFloatString(maxImg[1], 6);
      yMinEdit.Text := FormatFloatString(minImg[2], 6);
      yMaxEdit.Text := FormatFloatString(maxImg[2], 6);
    end;

    if vers =0 then
      stsImg.r2:= sqr(stsImg.r);

        {added in kFileVersion1, and sppressed afterwards}
    {if (vers = 1) and (mt=Mobius) then begin
      UpdatePositionCursors(True);
      for i:= 1 to 2 do
        for j:= 1 to 2 do
          for k:= 1 to 2 do
            crsEdit[i,j,k].text:= FormatFloatString(mm[i,k,k], 6);
    end;}

    alreadyDrawn := False;
    update;
    ChangeMapType(true);
    ReadAndDraw(nil);
    dirty:= False;
    title:= ExtractFileName(filename);
    anInt := length(title);
    if title[anInt - 3] = '.' then begin
      title := copy(title, 0, anInt - 4);
      caption:= title;
    end;
  end
end;

procedure TMapR2R2.Save(Sender : TObject);
begin
  if (not alreadyDrawn) or (not dirty) then
    Exit;
  if filename = '' then
    SaveAs(sender)
  else begin
    if filename[length(filename) - 3] <> '.' then  {add the extension}
      filename := filename + kFilter;
    WriteToDisk(filename);
  end;
end;

procedure TMapR2R2.SaveAs(Sender : TObject);
var
  Done : Boolean;
  lt : Integer;
  title : AnsiString;
begin
  if (not alreadyDrawn) then
    Exit;
  Done := False;
  while not Done do begin
    if SaveDialog1.Execute then begin
      filename := SaveDialog1.FileName;
      if filename[length(filename) - 3] <> '.' then
        filename := filename + kFilter;
        WriteToDisk(filename);
        dirty := False;
        title := ExtractFileName(filename);
        lt := length(title);
        caption := copy(title, 0, lt - 4);
        Done := True;
      end
      else
        Done := True;
    end;
end;

procedure TMapR2R2.SetBasicData;
var
  i: Integer;
begin
  for i:= 1 to 2 do begin
    edgesDom[i] := maxDom[i] - minDom[i];
    halfEdgesDom[i]:= edgesDom[i]/2;
    ctrDom[i] := (maxDom[i] + minDom[i]) / 2;
    edgesParam[i]:= maxParam[i]- minParam[i];
    if not wholeRange then begin
      edgesImg[i] := maxImg[i] - minImg[i];
      halfEdgesImg[i] := edgesImg[i]/2;
      ctrImg[i]:= (maxImg[i] + minImg[i]) / 2;
    end;
  end;
  SetCoordinateChangesDom(domRct, minDom, maxDom, preserveRatio);
  spanDom:= hypot(edgesDom[1], edgesDom[2]);
  radDom := spanDom /(sqrt(2)*kDomWidth);
end;

procedure TMapR2R2.SetClip(can: TCanvas; rWhole, rClip: TRect);
{Fill the whole 'pb'  with the color of the window, then clip to the
rectangle 'r' and then fill 'r' with white}
var
  rg: HRGN;
begin
  rg:= CreateRectRgn(rWhole.left, rWhole.top, rWhole.right, rWhole.bottom);
  SelectClipRgn(can.Handle, rg);
  can.Brush.Color:= color;
  can.FillRect(can.ClipRect);
  DeleteObject(rg);

  rg:= CreateRectRgn(rClip.left, rClip.top, rClip.right, rClip.bottom);
  SelectClipRgn(can.Handle, rg);
  can.Brush.Color:= clWhite;
  can.FillRect(can.ClipRect);
  DeleteObject(rg);
end;

procedure TMapR2R2.SetCoordinateChangesParam;
{set coefficients for the functions 'FromParamsPanelToParameters' and
 'FromParametersToParamPanel'}
var
  i: Integer;
begin
  for i:= 1 to 2 do
    panel2ParamSc[i]:= edgesParam[i]/kParamWidthMinus1;
end;

procedure TMapR2R2.SetUpComplexScene(fromFile: Boolean);
begin
  if not fromFile then begin
    prevSpot.x:= 100;
    prevSpot.Y:= 100;
    F1Edit.Text:= cwzText;
  end;
  thereIsLittleCircle:= False;
  xuvLabel.caption:= 'w(z)=';
  yuvLabel.visible:= False;
  uName.left:= 2;
  vName.left:= 2;
  uLabel.Left:= 192;
  vLabel.Left:= 192;
  uLabel.caption:= '< z.re <';
  vLabel.caption:= '< z.im <';
  xLabel.Left:= 590;
  yLabel.left:= 590;
  xLabel.caption:= '< w.re <';
  yLabel.caption:= '< w.im <';
  uName.caption:= 'z.re=';
  vName.caption:= 'z.im=';
  pName.left:= 2;
  qName.left:= 2;
  pName.Caption:= 'p.re=';
  qName.caption:= 'p.im=';
  pLabel.Left:= 120;
  qLabel.Left:= 120;
  pLabel.caption:= '<p.re<';
  qLabel.caption:= '<p.im<';
  xName.Left:= 338;
  yName.Left:= 338;
  xName.Caption:= 'w.re=';
  yName.caption:= 'w.im=';
end;

procedure TMapR2R2.SetUpRealScene(fromFile: Boolean);
begin
  if not fromFile then begin
    prevSpot.x:= 100;
    prevSpot.Y:= 100;
    F1Edit.Text:= cxuvText;
    F2Edit.text:= cyuvText;
  end;
  thereIsLittleCircle:= False;

  {adjust labels captions and positions}
  xuvLabel.caption:= 'x(u,v)=';
  yuvLabel.visible:= True;
  uName.left:= 12;
  vName.left:= 12;
  uLabel.Left:= 200;
  vLabel.Left:= 200;
  uLabel.caption:= '< u <';
  vLabel.caption:= '< v <';
  xLabel.Left:= 598;
  yLabel.left:= 598;
  xLabel.caption:= '< x <';
  yLabel.caption:= '< y <';
  uName.caption:= 'u =';
  vName.caption:= 'v =';
  pName.left:= 6;
  qName.left:= 6;
  pName.Caption:= 'p =';
  qName.caption:= 'q =';
  pLabel.Left:= 126;
  qLabel.Left:= 126;
  pLabel.caption:= '< p <';
  qLabel.caption:= '< q <';
  xName.Left:= 348;
  yName.Left:= 348;
  xName.Caption:= 'x =';
  yName.caption:= 'y =';
end;


procedure TMapR2R2.StoreSingLocusNode(currentX: Vector2; key: Integer);
begin
  inc(singLocus.last);
  if singLocus.last > High(singLocus.nodes) then
    SetLength(singLocus.nodes, singLocus.last+100);
  singLocus.nodes[singLocus.last].key:= key;
  singLocus.nodes[singLocus.last].x:= currentX;
  singLocus.nodes[singLocus.last].Pt:= DomToDomR3(currentX);
  singLocus.nodes[singLocus.last].imgPt:= DomToImgR3(currentX);
end;


procedure TMapR2R2.UpdateSingularLocus(typ: Byte);
var
  i: Integer;
  v: Vector2;
begin
  if typ = kDom then begin
    if singLocus.last > 0 then
      for i:= 0 to singLocus.last do begin
        v:= singLocus.nodes[i].x;
        singLocus.nodes[i].pt:= DomToDomR3(v);
      end;
  end
  else if singLocus.last > 0 then
    for i:= 0 to singLocus.last do begin
      v:= singLocus.nodes[i].x;
      singLocus.nodes[i].imgPt:= DomToImgR3(v);
    end;
end;

function TMapR2R2.VerifyImageBoundsEntries: Boolean;
begin
  try
    {if any of these readings fails, flow passes to "except" below
     and in general nothing happens}
    ReadConstantField(xMinEdit.Text,minImg[1]);
    ReadConstantField(xMaxEdit.Text,maxImg[1]);
    ReadConstantField(yMinEdit.Text,minImg[2]);
    ReadConstantField(yMaxEdit.Text,maxImg[2]);

    result:= (minImg[1] < maxImg[1]) and (minImg[2] < maxImg[2]);
  except on  E: Exception do
      result:= False;
  end;
end;

procedure TMapR2R2.WantToSave(Sender : TObject; var Action: TCloseAction);
{Asks the user if the results must be saved. If True, save them and close
 the program.}
var
  rslt : Boolean;
  aFilename : TFileName;
begin
  Action := caNone;
  if alreadyDrawn and dirty then
    begin
      SaveDialog1.Title := eWantToSave;
      rslt := SaveDialog1.Execute;
      if rslt then {save and exit}
        begin
          aFilename := SaveDialog1.FileName;
          WriteToDisk(filename);
          Action := caFree; {Exit the program}
        end
      else
        Action := caNone;
    end;
end;

(*   FROM THE VERSION kFileVersion0

procedure TMapR2R2.WriteToDisk(pathAndName: TFilename);
var
  mst: TMemoryStream;
  anInt, len, preferences: Integer;
  forFutureUses: Int64;
  str: AnsiString;
begin
  mst := TMemoryStream.Create;
   preferences:= cWholeRangeMask*ord(wholeRange) +
                cToDrawSingularLocusMask*ord(toDrawSingularLocus)+
                cStereographicMask*ord(modeStImage)+
                cPreserveRatioMask*ord(preserveRatio);

  {$O-} {There are some pitfalls if optimization is enabled. Disable it}

 //The meaning of the following code is straightforward
  try
    if pathAndName = '' then begin
      ReportIOError(-3);
      dirty := True;
    end
    else try
      mst.Seek(0, soFromBeginning);
      WriteStreamStr(mst,kCreator);
      WriteStreamStr(mst,kFilter);
      WriteStreamStr(mst,kFileVersion0);

      forFutureUses:= 0;
      mst.Write(forFutureUses, SizeOf(int64));

      anInt:= ord(mt);
      mst.Write(anInt, SizeOf(Integer));
      mst.Write(preferences, SizeOf(Integer));

      WriteStreamStr(mst, F1Edit.Text);
      WriteStreamStr(mst, F2Edit.Text);

      WriteStreamSomething(mst, minParam, SizeOf(Vector2));
      WriteStreamSomething(mst, maxParam, SizeOf(Vector2));
      mst.Write(p, SizeOf(Double));
      mst.Write(q, SizeOf(Double));
      WriteStreamSomething(mst, minDom, SizeOf(Vector2));
      WriteStreamSomething(mst, maxDom, SizeOf(Vector2));
      WriteStreamSomething(mst, minImg, SizeOf(Vector2));
      WriteStreamSomething(mst, maxImg, SizeOf(Vector2));
      WriteStreamSomething(mst, stsImg.c, SizeOf(Vector2));
      WriteStreamSomething(mst, stsImg.h, SizeOf(Vector3));
      WriteStreamSomething(mst, stsImg.v, SizeOf(Vector3));
      WriteStreamSomething(mst, stsImg.n, SizeOf(Vector3));
      mst.Write(stsImg.r, SizeOf(Double));
      dirty:= False;
      mst.SaveToFile(pathAndName);
    except
      dirty := True;
      ReportIOError(-4);
    end;

  finally
    mst.free;
    {$O+}{Enable optimizations}
  end;
end;*)


procedure TMapR2R2.WriteToDisk(pathAndName: TFilename);
var
  mst: TMemoryStream;
  anInt, preferences: Integer;
  forFutureUses: Int64;
begin
  mst := TMemoryStream.Create;
   preferences:= cWholeRangeMask*ord(wholeRange) +
                cToDrawSingularLocusMask*ord(toDrawSingularLocus)+
                cStImgMask*ord(modeStImage)+
                cPreserveRatioMask*ord(preserveRatio)+

                {added in the kFileVersion1}
                cStDomMask*ord(modeStDomain);

  {$O-} {There are some pitfalls if optimization is enabled. Disable it}

 //The meaning of the following code is straightforward


  try
    if pathAndName = '' then begin
      ReportIOError(-3);
      dirty := True;
    end
    else try
      mst.Seek(0, soFromBeginning);
      WriteStreamStr(mst,kCreator);
      WriteStreamStr(mst,kFilter);
      WriteStreamStr(mst,kFileVersion0);

      forFutureUses:= 0;
      mst.Write(forFutureUses, SizeOf(int64));

      anInt:= ord(mt);
      mst.Write(anInt, SizeOf(Integer));
      mst.Write(preferences, SizeOf(Integer));

      WriteStreamStr(mst, F1Edit.Text);
      WriteStreamStr(mst, F2Edit.Text);

      WriteStreamSomething(mst, minParam, SizeOf(Vector2));
      WriteStreamSomething(mst, maxParam, SizeOf(Vector2));
      mst.Write(p, SizeOf(Double));
      mst.Write(q, SizeOf(Double));
      WriteStreamSomething(mst, minDom, SizeOf(Vector2));
      WriteStreamSomething(mst, maxDom, SizeOf(Vector2));
      WriteStreamSomething(mst, minImg, SizeOf(Vector2));
      WriteStreamSomething(mst, maxImg, SizeOf(Vector2));
      WriteStreamSomething(mst, stsImg.c, SizeOf(Vector2));
      WriteStreamSomething(mst, stsImg.h, SizeOf(Vector3));
      WriteStreamSomething(mst, stsImg.v, SizeOf(Vector3));
      WriteStreamSomething(mst, stsImg.n, SizeOf(Vector3));
      mst.Write(stsImg.r, SizeOf(Double));

      {Added in fhe kFileVersion1}
      WriteStreamSomething(mst, stsDom, SizeOf(stsDom));
      WriteStreamSomething(mst, stsImg, SizeOf(stsImg));
      //WriteStreamSomething(mst, mm, SizeOf(mm));

      dirty:= False;
      mst.SaveToFile(pathAndName);
    except
      dirty := True;
      ReportIOError(-4);
    end;

  finally
    mst.free;
    {$O+}{Enable optimizations}
  end;
end;


end.

