{-----------------------------------------------------------------------------
 Unit Name: PrintDrawings
 Project Name : MapR2toR2
 Author:    Angel Montesinos Amilibia
 January  2009
 Purpose: Print texts and drawings
-----------------------------------------------------------------------------}
{

MapR2toR2, workshop for smooth maps from R^2 to R^2
Copyright (C) 2008-2009 Angel Montesinos Amilibia,
   University of Valencia, Spain


This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation,Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

You can email me to:
     montesin@uv.es

Or write me to:
     Angel Montesinos
     Departamento de Geometria y Topologia
     Vicente Andrs Estells, 1
     46100 Burjasot (Valencia)
     Spain

}


unit PrintDrawings;

interface

uses
  ConstTypesAndGlobals;

procedure SetPrinterRectangles;
procedure DrawOnPrinter;



implementation

uses
  Windows, Graphics, Types, Printers, Math, SysUtils,
  FMapR2R2, ResStrings, Maps;

var
  domRctPr, imgRctPr, usableR: TRect;
  pageW, pageH, titleTop, formulasTop: Integer;


procedure SetPrinterRectangles;
{Compute the rectangles in the printer canvas where must be printed
 the title, the drawings and the formulas. Set printer orientation}
var
  w: Integer;
  aux1, aux2, hLine: Integer;
begin
  Printer.Orientation:= poLandscape;
  pageW := Printer.PageWidth;
  pageH := Printer.PageHeight;

  {set margins}
  aux1 := pageW DIV 14;
  aux2 := pageH DIV 14;

  SetRect(usableR, aux1, aux2, pageW-aux1, pageH-aux2);
  w:= pageW-2*aux1;

  hLine := pageH DIV 40; {line height}

  {Set title top}
  titleTop:= usableR.Top;

  {Set formulas top}
  formulasTop:= usableR.Bottom - 3*hLine;

  {Set drawing rectangles}
  aux2:= (w*14) div 42;
  SetRect(domRctPr, usableR.Left, usableR.top + 4*hLine,
                  usableR.Left + aux2, usableR.top + 4*hLine + aux2);

  aux2:= Min((w*26) div 42, formulasTop - titleTop - hLine);

  SetRect(imgRctPr, usableR.right - aux2, titleTop, usableR.right,
           titleTop + aux2);
  formulasTop:= imgRctPr.Bottom + 2*hLine;
end;

procedure PrintTexts;
{Prints title, formulas, parameters, domain and image}
var
  can: TCanvas;
  separator : Char;
  t : AnsiString;
  oldCharSet : TFontCharSet;
  oldPen : TPen;
  x, y, hLine: Integer;
begin
  can:= Printer.Canvas;
  separator := DecimalSeparator;
  DecimalSeparator := '.';
  oldPen := can.Pen;
  hLine:= pageH div 35;
  can.Font.Name := 'Arial';
  oldCharSet := can.Font.charset;
  can.Font.charset := 1;
  can.Font.Style := [fsBold];
  can.Font.Height := (hLine*3) div 4;
  if MapR2R2.mt = Real then
    t := 'MapR2R2 :  '+ eRealMap
  else if MapR2R2.mt = Complex then
    t := 'MapR2R2 :  '+ eComplexMap;

  x := usableR.left;
  y := titleTop;
  try
    can.TextOut(x, y, t); {print general title}
    {Print below specific subtitle}
    can.Font.Height:= (hLine*4) div 5;
    if MapR2R2.filename= '' then
      t:= eUntitled
    else
      t:= ExtracTitle(MapR2R2.filename, kFilter);
    y:= y + hLine;
    can.TextOut(x, y, t);

    y:= formulasTop;
    can.Font.Style := [];
    can.Font.Height:= (hLine*4) div 5;
    if MapR2R2.mt = Real then
      t:= 'x(u,v) = ' + Trim(MapR2R2.F1Edit.Text)
    else if MapR2R2.mt = Complex then
      t:= 'w(z) = ' + Trim(MapR2R2.F1Edit.Text)
    else
      t:= 'w(z) = (a z + b)/(c z + d)';
    x:= (usableR.left + usableR.right - can.TextWidth(t)) div 2;
    can.TextOut(x, y, t); {print first formula}

    if MapR2R2.mt = Real then begin
      y:= y + hLine;
      t:= 'y(u,v) = ' + Trim(MapR2R2.F2Edit.Text);
      x:= (usableR.left + usableR.right - can.TextWidth(t)) div 2;
      can.TextOut(x, y, t); {print second formula}
    end;

    hLine:= pageH div 40;
    can.Font.Height := (hLine*4) div 5;

    x:= domRctPr.Left;
    with MapR2R2 do begin
      y:= domRctPr.Bottom + 2*hLine;
      if mt = Real then
        t:= 'p = ' +  pValue.text + ' ,   q = ' + qValue.text
      else if mt = Complex then
        t:= 'Re(p) = ' +  pValue.text + ' ,   Im(p) = ' + qValue.text ;
      can.TextOut(x, y, t); {print parameter values}
      y:= y + 2*hLine;
      if mt = Real then
        t:= uMinEdit.text + ' < u < ' + uMaxEdit.text + ' ,   ' +
            vMinEdit.text + ' < v < ' + vMaxEdit.text
      else
        t:= uMinEdit.text + ' < Re(z) < ' + uMaxEdit.text + ' ,   ' +
            vMinEdit.text + ' < Im(z) < ' + vMaxEdit.text;
      can.TextOut(x, y, t); {print independent variables limits}

      y:= y + 2*hLine;
      if mt = Real then
        t:= FormatFloatString(minImg[1], 6) + ' < x < ' + FormatFloatString(maxImg[1], 6)
        + ' ,   '
        + FormatFloatString(minImg[2], 6) + ' < y < ' + FormatFloatString(maxImg[2], 6)
      else
        t:= FormatFloatString(minImg[1], 6) + ' < Re(w) < ' + FormatFloatString(maxImg[1], 6)
        + ' ,   '
        + FormatFloatString(minImg[2], 6) + ' < Im(w) < ' + FormatFloatString(maxImg[2], 6);
      can.TextOut(x, y, t); {print independent variables limits};
    end;

  finally
    DecimalSeparator:= separator;
    can.pen:= oldPen;
    can.Font.charset:= oldCharSet;
  end;
end;


procedure DrawOnPrinter;
var
  r: TRect;
  rg: HRGN;
begin
  with printer.canvas do begin
    pen.Width:= 2;
    pen.color:= $00777777;
    moveto(imgRctPr.left,imgRctPr.top);
    lineto(imgRctPr.right,imgRctPr.top);
    lineto(imgRctPr.right,imgRctPr.bottom);
    lineto(imgRctPr.left,imgRctPr.bottom);
    lineto(imgRctPr.left,imgRctPr.top);
  end;
  PrintTexts;

  with MapR2R2 do begin
    {clip to the actual rectangle where the drawing of domain is to be done}
    rg:= CreateRectRgnIndirect(domRctPr);
    SelectClipRgn(Printer.canvas.Handle, rg);
    DeleteObject(rg);
    try
      SetCoordinateChangesDom(domRctPr, minDom, maxDom,preserveRatio);
      DrawDomainCurves(domRctPr, printer.canvas);
      if toDrawSingularLocus then
        DrawSingularLocusDom(domRctPr, printer.canvas);

      {clip to the actual rectangle where the drawing of image is to be done}
      rg:= CreateRectRgnIndirect(imgRctPr);
      SelectClipRgn(Printer.canvas.Handle, rg);
      DeleteObject(rg);

      SetCoordinateChangesImg(imgRctPr, minImg, maxImg, preserveRatio);
      DrawImageCurves(imgRctPr, printer.canvas);
      if toDrawSingularLocus then
        DrawSingularLocusImg(imgRctPr, printer.canvas);

      {restore the clip region to maximum size}
      SetRect(r,0,0,Printer.pageWidth, Printer.pageHeight);
      rg:= CreateRectRgnIndirect(r);
      SelectClipRgn(Printer.canvas.Handle, rg);
      DeleteObject(rg);
    finally
      SetCoordinateChangesDom(domRct, minDom, maxDom,preserveRatio);
      SetCoordinateChangesImg(imgRct, minDom, maxDom,preserveRatio);
    end;
  end;
end;

end.
